from .constants import DISSOLVED, IN_FOLLOWUP, RECRUITING
from .screening import (
    INVALID_APPOINTMENT_DATE,
    INVALID_CHANGE_ALREADY_SCREENED,
    INVALID_GROUP,
    ConsentRefusalFormValidator,
    HealthFacilityFormValidator,
    HealthTalkLogFormValidator,
    PatientCallFormValidator,
    PatientGroupFormValidator,
    PatientLogFormValidator,
    SubjectScreeningFormValidator,
)
from .subject import (
    DmReviewFormValidator,
    DrugRefillDmFormValidator,
    DrugRefillHivFormValidator,
    DrugRefillHtnFormValidator,
    HealthEconomicsFormValidator,
    HivReviewFormValidator,
    HtnReviewFormValidator,
    SocialHarmsFormValidator,
)



#!/usr/bin/env python
import logging
import os.path
import sys
from pathlib import Path

import django
from django.conf import settings
from django.test.runner import DiscoverRunner

app_name = "intecomm_form_validators"
base_dir = Path(__file__).resolve().parent.parent

DEFAULT_SETTINGS = dict(  # nosec B106
    BASE_DIR=Path(__file__).resolve().parent.parent,
    SECRET_KEY="django-insecure",  # nosec B106
    DEBUG=True,
    SUBJECT_CONSENT_MODEL=None,
    SUBJECT_SCREENING_MODEL=None,
    EDC_DX_LABELS=dict(hiv="HIV", dm="Diabetes", htn="Hypertension"),
    ALLOWED_HOSTS=[],
    INSTALLED_APPS=[
        "django.contrib.admin",
        "django.contrib.auth",
        "django.contrib.sites",
        "django.contrib.contenttypes",
        "django.contrib.sessions",
        "django.contrib.messages",
        "django.contrib.staticfiles",
        "intecomm_form_validators.apps.AppConfig",
    ],
    MIDDLEWARE=[
        "django.middleware.security.SecurityMiddleware",
        "django.contrib.sessions.middleware.SessionMiddleware",
        "django.middleware.common.CommonMiddleware",
        "django.middleware.csrf.CsrfViewMiddleware",
        "django.contrib.auth.middleware.AuthenticationMiddleware",
        "django.contrib.messages.middleware.MessageMiddleware",
        "django.middleware.clickjacking.XFrameOptionsMiddleware",
        "edc_dashboard.middleware.DashboardMiddleware",
    ],
    ROOT_URLCONF="intecomm_form_validators.urls",
    TEMPLATES=[
        {
            "BACKEND": "django.template.backends.django.DjangoTemplates",
            "DIRS": [],
            "APP_DIRS": True,
            "OPTIONS": {
                "context_processors": [
                    "django.template.context_processors.debug",
                    "django.template.context_processors.request",
                    "django.contrib.auth.context_processors.auth",
                    "django.contrib.messages.context_processors.messages",
                ],
            },
        },
    ],
    WSGI_APPLICATION="intecomm_form_validators.wsgi.application",
    DATABASES={
        "default": {
            "ENGINE": "django.db.backends.sqlite3",
            "NAME": os.path.join(base_dir, "db.sqlite3"),
        }
    },
    AUTH_PASSWORD_VALIDATORS=[
        {
            "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
        },
        {
            "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
        },
        {
            "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
        },
        {
            "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
        },
    ],
    LANGUAGE_CODE="en-us",
    TIME_ZONE="UTC",
    USE_I18N=True,
    USE_L10N=True,
    USE_TZ=True,
    STATIC_URL="/static/",
    DEFAULT_AUTO_FIELD="django.db.models.BigAutoField",
)


def main():
    if not settings.configured:
        settings.configure(**DEFAULT_SETTINGS)
    django.setup()
    tags = [t.split("=")[1] for t in sys.argv if t.startswith("--tag")]
    failfast = True if [t for t in sys.argv if t == "--failfast"] else False
    failures = DiscoverRunner(failfast=failfast, tags=tags).run_tests([])
    sys.exit(bool(failures))


if __name__ == "__main__":
    logging.basicConfig()
    main()



from dateutil.relativedelta import relativedelta
from django.utils.html import format_html
from edc_constants.constants import DM, HIV, HTN, MALE, NO, YES
from edc_form_validators import INVALID_ERROR, FormValidator
from edc_model import InvalidFormat, duration_to_date
from edc_screening.form_validator_mixins import SubjectScreeningFormValidatorMixin

INVALID_DURATION_IN_CARE = "invalid_duration_in_care"


class SubjectScreeningFormValidator(SubjectScreeningFormValidatorMixin, FormValidator):
    def clean(self):
        self.validate_stable_in_care_on_patient_log()
        self.validate_health_talks_on_patient_log()
        self.get_consent_for_period_or_raise()
        self.validate_gender_against_patient_log()
        self.validate_age_in_years_against_patient_log()

        if (
            self.cleaned_data.get("consent_ability")
            and self.cleaned_data.get("consent_ability") == NO
        ):
            self.raise_validation_error(
                {
                    "consent_ability": (
                        "You may NOT screen this subject without their verbal consent."
                    )
                },
                INVALID_ERROR,
            )

        self.required_if(YES, field="in_care_6m", field_required="in_care_duration")
        self.duration_in_care_is_6m_or_more_or_raise()

        self.validate_hiv_section()
        self.validate_dm_section()
        self.validate_htn_section()

        self.not_applicable_if(
            MALE, field="gender", field_applicable="pregnant", inverse=False
        )

        self.validate_suitability_for_study()

    def validate_stable_in_care_on_patient_log(self):
        if self.patient_log.stable != YES:
            link = (
                f'<a href="{self.patient_log.get_changelist_url()}?'
                f'q={str(self.patient_log.id)}">{self.patient_log}</a>'
            )
            errmsg = format_html(
                "Invalid. Patient is not known to be stable and in-care. "
                f"See patient log for {link}."
            )
            self.raise_validation_error(errmsg, error_code=INVALID_ERROR)

        pass

    def validate_health_talks_on_patient_log(self) -> bool:
        if self.patient_log.first_health_talk not in [YES, NO]:
            link = (
                f'<a href="{self.patient_log.get_changelist_url()}?'
                f'q={str(self.patient_log.id)}">{self.patient_log}</a>'
            )
            errmsg = format_html(
                "Invalid. Has patient attended the first health talk? "
                f"See patient log for {link}."
            )
            self.raise_validation_error(errmsg, error_code=INVALID_ERROR)
        elif self.patient_log.second_health_talk not in [YES, NO]:
            link = (
                f'<a href="{self.patient_log.get_changelist_url()}?'
                f'q={str(self.patient_log.id)}">{self.patient_log}</a>'
            )
            errmsg = format_html(
                "Invalid. Has patient attended the second health talk? "
                f"See patient log for {link}."
            )
            self.raise_validation_error(errmsg, error_code=INVALID_ERROR)
        return True

    def validate_gender_against_patient_log(self):
        if self.cleaned_data.get("gender") != self.patient_log.gender:
            self.raise_validation_error(
                {
                    "gender": (
                        f"Invalid. Expected {self.patient_log.get_gender_display()}. "
                        "See Patient Log."
                    )
                },
                INVALID_ERROR,
            )

        pass

    def validate_age_in_years_against_patient_log(self):
        if self.cleaned_data.get("age_in_years") != self.patient_log.age_in_years:
            self.raise_validation_error(
                {
                    "age_in_years": (
                        f"Invalid. Expected {self.patient_log.age_in_years}. "
                        "See Patient Log."
                    )
                },
                INVALID_ERROR,
            )

        pass

    def duration_in_care_is_6m_or_more_or_raise(self, fieldname: str = None) -> None:
        fieldname = fieldname or "in_care_duration"
        in_care_duration = self.cleaned_data.get(fieldname)
        report_datetime = self.cleaned_data.get("report_datetime")
        if report_datetime and in_care_duration:
            try:
                dt = duration_to_date(in_care_duration, report_datetime)
            except InvalidFormat as e:
                self.raise_validation_error({fieldname: f"Invalid format. {e}"}, INVALID_ERROR)
            if dt + relativedelta(months=6) > report_datetime:
                self.raise_validation_error(
                    {fieldname: "Expected at least 6m from the report date"},
                    INVALID_DURATION_IN_CARE,
                )

    def validate_hiv_section(self):
        self.validate_condition(HIV, "hiv_dx")
        self.applicable_if(YES, field="hiv_dx", field_applicable="hiv_dx_6m")
        self.required_if(YES, field="hiv_dx_6m", field_required="hiv_dx_ago")
        self.duration_in_care_is_6m_or_more_or_raise("hiv_dx_ago")
        self.applicable_if(YES, field="hiv_dx", field_applicable="art_unchanged_3m")
        self.applicable_if(YES, field="hiv_dx", field_applicable="art_stable")
        self.applicable_if(YES, field="hiv_dx", field_applicable="art_adherent")

    def validate_dm_section(self):
        self.validate_condition(DM, "dm_dx")
        self.applicable_if(YES, field="dm_dx", field_applicable="dm_dx_6m")
        self.required_if(YES, field="dm_dx_6m", field_required="dm_dx_ago")
        self.duration_in_care_is_6m_or_more_or_raise("dm_dx_ago")
        self.applicable_if(YES, field="dm_dx", field_applicable="dm_complications")

    def validate_htn_section(self):
        self.validate_condition(HTN, "htn_dx")
        self.applicable_if(YES, field="htn_dx", field_applicable="htn_dx_6m")
        self.required_if(YES, field="htn_dx_6m", field_required="htn_dx_ago")
        self.duration_in_care_is_6m_or_more_or_raise("htn_dx_ago")
        self.applicable_if(YES, field="htn_dx", field_applicable="htn_complications")

    def validate_condition(self, name, field):
        conditions = self.patient_log.conditions
        if not self.patient_log.conditions:
            self.raise_validation_error(
                "No conditions (HIV/DM/HTN) have been indicated for this patient. See "
                "the Patient Log",
                INVALID_ERROR,
            )
        else:
            if not conditions.filter(name=name) and self.cleaned_data.get(field) == YES:
                self.raise_validation_error(
                    {
                        field: (
                            f"Invalid. {name.upper()} was not indicated "
                            "as a condition on the Patient Log"
                        ),
                    },
                    INVALID_ERROR,
                )
            elif conditions.filter(name=name) and self.cleaned_data.get(field) == NO:
                self.raise_validation_error(
                    {
                        field: (
                            f"Invalid. {name.upper()} was indicated "
                            "as a condition on the Patient Log"
                        ),
                    },
                    INVALID_ERROR,
                )

    def validate_suitability_for_study(self):
        self.required_if(
            YES, field="unsuitable_for_study", field_required="reasons_unsuitable"
        )
        self.applicable_if(
            YES, field="unsuitable_for_study", field_applicable="unsuitable_agreed"
        )
        if self.cleaned_data.get("unsuitable_agreed") == NO:
            self.raise_validation_error(
                {
                    "unsuitable_agreed": "The study coordinator MUST agree "
                    "with your assessment. Please discuss before continuing."
                },
                INVALID_ERROR,
            )

    @property
    def patient_log(self):
        if patient_log := self.cleaned_data.get("patient_log"):
            return patient_log
        return self.instance.patient_log



from edc_adherence.form_validator_mixin import MedicationAdherenceFormValidatorMixin
from edc_crf.crf_form_validator_mixins import CrfFormValidatorMixin
from edc_form_validators import FormValidator


class HtnMedicationAdherenceFormValidator(
    MedicationAdherenceFormValidatorMixin,
    CrfFormValidatorMixin,
    FormValidator,
):
    pass



from edc_constants.constants import YES
from edc_crf.crf_form_validator_mixins import CrfFormValidatorMixin
from edc_dx_review.utils import raise_if_clinical_review_does_not_exist
from edc_form_validators import FormValidator


class ComplicationsBaselineFormValidator(CrfFormValidatorMixin, FormValidator):
    def clean(self):
        raise_if_clinical_review_does_not_exist(self.cleaned_data.get("subject_visit"))
        self.required_if(YES, field="stroke", field_required="stroke_ago")
        # self.estimated_date_from_ago("stroke_ago")
        self.required_if(YES, field="heart_attack", field_required="heart_attack_ago")
        # self.estimated_date_from_ago("heart_attack_ago")
        self.required_if(YES, field="renal_disease", field_required="renal_disease_ago")
        # self.estimated_date_from_ago("renal_disease_ago")
        self.required_if(YES, field="vision", field_required="vision_ago")
        # self.estimated_date_from_ago("vision_ago")
        self.required_if(YES, field="numbness", field_required="numbness_ago")
        # self.estimated_date_from_ago("numbness_ago")
        self.required_if(YES, field="foot_ulcers", field_required="foot_ulcers_ago")
        # self.estimated_date_from_ago("foot_ulcers_ago")
        self.required_if(YES, field="complications", field_required="complications_other")



from django.utils.html import format_html
from edc_constants.constants import COMPLETE, NEW, YES
from edc_form_validators import FormValidator

from ..constants import RECRUITING
from ..utils import (
    PatientGroupRatioError,
    PatientGroupSizeError,
    PatientNotConsentedError,
    PatientNotScreenedError,
    PatientNotStableError,
    confirm_patient_group_size_or_raise,
    confirm_patients_stable_and_screened_and_consented_or_raise,
    get_group_size_for_ratio,
    verify_patient_group_ratio_raise,
)

INVALID_PATIENT_COUNT = "INVALID_PATIENT_COUNT"
INVALID_RANDOMIZE = "INVALID_RANDOMIZE"
INVALID_PATIENT = "INVALID_PATIENT"
INVALID_CONDITION_RATIO = "INVALID_CONDITION_RATIO"
INVALID_STATUS = "INVALID_STATUS"


class PatientGroupFormValidator(FormValidator):
    def clean(self):
        self.block_changes_if_already_randomized()

        if self.cleaned_data.get("status") and self.cleaned_data.get("status") not in [
            NEW,
            RECRUITING,
            COMPLETE,
        ]:
            self.raise_validation_error({"status": "Invalid selection"}, INVALID_STATUS)

        if (
            self.cleaned_data.get("status") != COMPLETE
            and self.cleaned_data.get("randomize_now") == YES
        ):
            self.raise_validation_error(
                {"randomize_now": "Invalid. Group is not complete"}, INVALID_RANDOMIZE
            )

        if self.cleaned_data.get("status") == COMPLETE:
            self.confirm_patients_stable_and_screened_and_consented_or_raise()
            self.confirm_patient_group_size_or_raise()
            self.verify_patient_group_ratio_raise()

        if (
            self.cleaned_data.get("randomize_now") == YES
            and self.cleaned_data.get("confirm_randomize_now") != "RANDOMIZE"
        ):
            self.raise_validation_error(
                {
                    "confirm_randomize_now": (
                        "If you wish to randomize this group, please confirm"
                    )
                },
                INVALID_RANDOMIZE,
            )
        self.not_required_if_true(
            self.cleaned_data.get("randomize_now") != YES,
            field="confirm_randomize_now",
            msg="Only complete if you are ready to randomize now.",
        )

    def block_changes_if_already_randomized(self):
        if self.instance.randomized:
            self.raise_validation_error(
                "A randomized group may not be changed", INVALID_RANDOMIZE
            )

    def confirm_patient_group_size_or_raise(self):
        try:
            confirm_patient_group_size_or_raise(
                bypass_group_size_min=self.cleaned_data.get("bypass_group_size_min"),
                patients=self.cleaned_data.get("patients") or self.instance.patients,
            )
        except PatientGroupSizeError as e:
            self.raise_validation_error({"__all__": str(e)}, INVALID_PATIENT_COUNT)

    def confirm_patients_stable_and_screened_and_consented_or_raise(self):
        try:
            confirm_patients_stable_and_screened_and_consented_or_raise(
                patients=self.cleaned_data.get("patients") or self.instance.patients
            )
        except (PatientNotStableError, PatientNotScreenedError, PatientNotConsentedError) as e:
            self.raise_validation_error({"__all__": str(e)}, INVALID_PATIENT)

    def verify_patient_group_ratio_raise(self):
        patients = self.cleaned_data.get("patients") or self.instance.patients
        if (
            not self.cleaned_data.get("bypass_group_ratio")
            and patients.all().count() >= get_group_size_for_ratio()
        ):
            try:
                verify_patient_group_ratio_raise(patients=patients.all())
            except PatientGroupRatioError as e:
                group_name = self.cleaned_data.get("name")
                errmsg = (
                    f'See group <a href="{self.instance.get_changelist_url(group_name)}">'
                    f"{group_name}</a>"
                )
                self.raise_validation_error(
                    {"__all__": format_html(f"{e} {errmsg}")}, INVALID_CONDITION_RATIO
                )

    def block_changes_if_randomized(self):
        if self.instance.randomized:
            self.raise_validation_error(
                "A randomized group may not be changed", INVALID_RANDOMIZE
            )



from __future__ import annotations

from django import forms
from django_mock_queries.query import MockSet
from edc_constants.constants import COMPLETE, NO, YES

from intecomm_form_validators.constants import RECRUITING
from intecomm_form_validators.screening import PatientGroupFormValidator as Base

from ..mock_models import PatientGroupMockModel
from ..test_case_mixin import TestCaseMixin


class PatientGroupTests(TestCaseMixin):
    @staticmethod
    def get_form_validator_cls(subject_screening=None):
        class PatientGroupFormValidator(Base):
            @property
            def subject_screening(self):
                return subject_screening

        return PatientGroupFormValidator

    def test_raises_if_randomized(self):
        patients = self.get_mock_patients(dm=10, htn=0, hiv=4)
        patient_group = PatientGroupMockModel(randomized=True, patients=MockSet(*patients))
        form_validator = self.get_form_validator_cls()(
            cleaned_data={}, instance=patient_group, model=PatientGroupMockModel
        )
        self.assertRaises(forms.ValidationError, form_validator.validate)
        patient_group = PatientGroupMockModel(randomized=False, patients=MockSet(*patients))
        form_validator = self.get_form_validator_cls()(
            cleaned_data={}, instance=patient_group, model=PatientGroupMockModel
        )
        try:
            form_validator.validate()
        except forms.ValidationError:
            self.fail("ValidationError unexpectedly raised")

    def test_raises_if_status_not_complete(self):
        patients = self.get_mock_patients(dm=10, htn=0, hiv=4)
        patient_group = PatientGroupMockModel()
        form_validator = self.get_form_validator_cls()(
            cleaned_data=dict(
                status=RECRUITING, randomize_now=YES, patients=MockSet(*patients)
            ),
            instance=patient_group,
            model=PatientGroupMockModel,
        )
        self.assertRaises(forms.ValidationError, form_validator.validate)
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("Invalid. Group is not complete", cm.exception.messages)

    def test_group_size_not_ok(self):
        patients = self.get_mock_patients(
            dm=10, htn=0, hiv=2, stable=True, screen=True, consent=True
        )
        patient_group = PatientGroupMockModel()
        form_validator = self.get_form_validator_cls()(
            cleaned_data={
                "status": COMPLETE,
                "randomize_now": NO,
                "patients": MockSet(*patients),
            },
            instance=patient_group,
            model=PatientGroupMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn(
            "Patient group must have at least 14 patients. Got 12.", cm.exception.messages
        )

    def test_group_size_ok(self):
        patients = self.get_mock_patients(
            dm=10, htn=0, hiv=4, stable=True, screen=True, consent=True
        )
        patient_group = PatientGroupMockModel(randomized=False, patients=MockSet(*patients))
        form_validator = self.get_form_validator_cls()(
            cleaned_data={
                "status": COMPLETE,
                "randomize_now": NO,
                "patients": MockSet(*patients),
            },
            instance=patient_group,
            model=PatientGroupMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError:
            self.fail("ValidationError unexpectedly raised")

    def test_group_size_overridden(self):
        patients = self.get_mock_patients(dm=10, htn=0, hiv=4)
        patient_group = PatientGroupMockModel(randomized=False, patients=MockSet(*patients))
        form_validator = self.get_form_validator_cls()(
            cleaned_data={
                "status": COMPLETE,
                "randomize_now": NO,
                "patients": MockSet(*patients),
            },
            instance=patient_group,
            model=PatientGroupMockModel,
        )
        # gets past group size check and starts reviewing patients
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn(
            "Patient is not known to be stable and in-care", "|".join(cm.exception.messages)
        )

    def test_group_size_too_small(self):
        patients = self.get_mock_patients(dm=10, htn=0, hiv=3)
        form_validator = self.get_form_validator_cls()(
            cleaned_data=dict(patients=MockSet(*patients), status=COMPLETE),
            instance=PatientGroupMockModel(randomized=False),
            model=PatientGroupMockModel,
        )
        self.assertRaises(forms.ValidationError, form_validator.validate)

    def test_review_patients_in_group_none_stable(self):
        patients = self.get_mock_patients(dm=10, htn=0, hiv=4, stable=False)
        patient_group = PatientGroupMockModel(randomized=False, patients=MockSet(*patients))
        form_validator = self.get_form_validator_cls()(
            cleaned_data={
                "status": COMPLETE,
                "randomize_now": NO,
                "patients": MockSet(*patients),
            },
            instance=patient_group,
            model=PatientGroupMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn(
            "Patient is not known to be stable and in-care", "|".join(cm.exception.messages)
        )

    def test_review_patients_in_group_all_stable(self):
        patients = self.get_mock_patients(dm=10, htn=0, hiv=4, stable=YES)
        patient_group = PatientGroupMockModel(randomized=False, patients=MockSet(*patients))
        form_validator = self.get_form_validator_cls()(
            cleaned_data={
                "status": COMPLETE,
                "randomize_now": NO,
                "patients": MockSet(*patients),
            },
            instance=patient_group,
            model=PatientGroupMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn(
            "Patient has not screened for eligibility", "|".join(cm.exception.messages)
        )

    def test_review_patients_in_group_all_screened(self):
        patients = self.get_mock_patients(dm=10, htn=0, hiv=4, stable=YES, screen=True)
        patient_group = PatientGroupMockModel()
        form_validator = self.get_form_validator_cls()(
            cleaned_data={
                "status": COMPLETE,
                "randomize_now": NO,
                "patients": MockSet(*patients),
            },
            instance=patient_group,
            model=PatientGroupMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("Patient has not consented", "|".join(cm.exception.messages))

    def test_review_patients_in_group_all_consented(self):
        patients = self.get_mock_patients(
            dm=10, htn=0, hiv=4, stable=YES, screen=True, consent=True
        )
        patient_group = PatientGroupMockModel(randomized=False, patients=MockSet(*patients))
        form_validator = self.get_form_validator_cls()(
            cleaned_data={
                "status": COMPLETE,
                "randomize_now": NO,
                "patients": MockSet(*patients),
            },
            instance=patient_group,
            model=PatientGroupMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError:
            self.fail("ValidationError unexpectedly raised")

    def test_ratio_ok(self):
        patients = self.get_mock_patients(
            dm=10, htn=0, hiv=4, stable=YES, screen=True, consent=True
        )
        patient_group = PatientGroupMockModel(randomized=False, patients=MockSet(*patients))
        form_validator = self.get_form_validator_cls()(
            cleaned_data={
                "status": COMPLETE,
                "randomize_now": NO,
                "patients": MockSet(*patients),
            },
            instance=patient_group,
            model=PatientGroupMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError:
            self.fail("ValidationError unexpectedly raised")

    def test_ratio_ok2(self):
        patients = self.get_mock_patients(
            dm=4, htn=4, hiv=4, hiv_ncd=2, stable=YES, screen=True, consent=True
        )
        patient_group = PatientGroupMockModel(randomized=False, patients=MockSet(*patients))
        form_validator = self.get_form_validator_cls()(
            cleaned_data={
                "status": COMPLETE,
                "randomize_now": NO,
                "patients": MockSet(*patients),
            },
            instance=patient_group,
            model=PatientGroupMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError:
            self.fail("ValidationError unexpectedly raised")

    def test_ratio_not_ok(self):
        for ratio in [[10, 0, 6], [11, 0, 3], [12, 0, 7], [13, 0, 7]]:
            with self.subTest(ratio=ratio):
                dm, htn, hiv = ratio
                patients = self.get_mock_patients(
                    dm, htn, hiv, stable=YES, screen=True, consent=True
                )
                patient_group = PatientGroupMockModel(patients=MockSet(*patients))
                form_validator = self.get_form_validator_cls()(
                    cleaned_data={
                        "status": COMPLETE,
                        "randomize": NO,
                        "patients": MockSet(*patients),
                    },
                    instance=patient_group,
                    model=PatientGroupMockModel,
                )
                with self.assertRaises(forms.ValidationError) as cm:
                    form_validator.validate()
                self.assertIn("Ratio NDC:HIV not met", "|".join(cm.exception.messages))

    def test_confirm_randomize_now_required_if_randomize_now_yes(self):
        patients = self.get_mock_patients(
            dm=10, htn=0, hiv=4, stable=YES, screen=True, consent=True
        )
        patient_group = PatientGroupMockModel(randomized=False, patients=MockSet(*patients))

        form_validator = self.get_form_validator_cls()(
            cleaned_data={
                "status": COMPLETE,
                "randomize_now": YES,
                "confirm_randomize_now": "",
                "patients": MockSet(*patients),
            },
            instance=patient_group,
            model=PatientGroupMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn(
            "If you wish to randomize this group, please confirm",
            str(cm.exception.error_dict.get("confirm_randomize_now")),
        )

    def test_confirm_randomize_now_not_required_if_randomize_now_not_yes(self):
        patients = self.get_mock_patients(
            dm=10, htn=0, hiv=4, stable=YES, screen=True, consent=True
        )
        patient_group = PatientGroupMockModel(randomized=False, patients=MockSet(*patients))

        for randomize_now in [NO, "xxx"]:
            for confirm_randomize_now in ["R", "Rand", "RANDOMIZE"]:
                with self.subTest(
                    randomize_now=randomize_now,
                    confirm_randomize_now=confirm_randomize_now,
                ):
                    form_validator = self.get_form_validator_cls()(
                        cleaned_data={
                            "status": COMPLETE,
                            "randomize_now": randomize_now,
                            "confirm_randomize_now": confirm_randomize_now,
                            "patients": MockSet(*patients),
                        },
                        instance=patient_group,
                        model=PatientGroupMockModel,
                    )
                    with self.assertRaises(forms.ValidationError) as cm:
                        form_validator.validate()
                    self.assertIn(
                        "This field is not required. "
                        "Only complete if you are ready to randomize now.",
                        str(cm.exception.error_dict.get("confirm_randomize_now")),
                    )

    def test_confirm_randomize_now_RANDOMIZE_randomize_now_yes_ok(self):
        patients = self.get_mock_patients(
            dm=10, htn=0, hiv=4, stable=YES, screen=True, consent=True
        )
        patient_group = PatientGroupMockModel(randomized=False, patients=MockSet(*patients))

        form_validator = self.get_form_validator_cls()(
            cleaned_data={
                "status": COMPLETE,
                "randomize_now": YES,
                "confirm_randomize_now": "RANDOMIZE",
                "patients": MockSet(*patients),
            },
            instance=patient_group,
            model=PatientGroupMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError:
            self.fail("ValidationError unexpectedly raised")



from django import forms
from django.test import TestCase
from django_mock_queries.query import MockModel, MockSet
from edc_constants.constants import DM, FEMALE, MALE, NO, NOT_APPLICABLE, TBD, YES
from edc_utils import get_utcnow

from intecomm_form_validators.screening import SubjectScreeningFormValidator as Base


class SubjectScreeningMockModel(MockModel):
    def __init__(self, *args, **kwargs):
        kwargs["mock_name"] = "SubjectScreening"
        super().__init__(*args, **kwargs)


class PatientGroupMockModel(MockModel):
    def __init__(self, *args, **kwargs):
        kwargs["mock_name"] = "PatientGroup"
        super().__init__(*args, **kwargs)


class PatientLogMockModel(MockModel):
    def __init__(self, *args, **kwargs):
        kwargs["mock_name"] = "PatientLog"
        super().__init__(*args, **kwargs)

    def get_gender_display(self):
        return "MALE" if self.gender == MALE else "FEMALE"

    def get_changelist_url(self):
        return "some_url"


class SubjectScreeningTests(TestCase):
    @staticmethod
    def get_form_validator_cls(subject_screening=None):
        class SubjectScreeningFormValidator(Base):
            def get_consent_for_period_or_raise(self):
                return None

        return SubjectScreeningFormValidator

    def patient_log(self, **kwargs):
        """Default is stable not screened"""
        opts = dict(
            name="THING ONE",
            gender=MALE,
            stable=YES,
            screening_identifier=None,
            subject_identifier=None,
            conditions=MockSet(
                MockModel(
                    mock_name="Conditions",
                    name=DM,
                )
            ),
            first_health_talk=NO,
            second_health_talk=NO,
        )
        opts.update(**kwargs)
        return PatientLogMockModel(**opts)

    def get_cleaned_data(self):
        return {
            "gender": MALE,
            "patient_log": self.patient_log(gender=MALE),
            "consent_ability": YES,
            "in_care_6m": YES,
            "in_care_duration": "5y",
            "hiv_dx": NO,
            "dm_dx": YES,
            "htn_dx": NO,
        }

    def test_cleaned_data_ok(self):
        cleaned_data = self.get_cleaned_data()
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=SubjectScreeningMockModel(),
            model=SubjectScreeningMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_gender(self):
        form_validator = self.get_form_validator_cls()(
            cleaned_data={"gender": MALE, "patient_log": self.patient_log(gender=FEMALE)},
            instance=SubjectScreeningMockModel(),
            model=SubjectScreeningMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("Invalid. Expected FEMALE", "|".join(cm.exception.messages))

    def test_gender_not_matching(self):
        form_validator = self.get_form_validator_cls()(
            cleaned_data={"gender": MALE, "patient_log": self.patient_log(gender=FEMALE)},
            instance=SubjectScreeningMockModel(),
            model=SubjectScreeningMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("Invalid. Expected FEMALE", "|".join(cm.exception.messages))

    def test_consent_ability(self):
        form_validator = self.get_form_validator_cls()(
            cleaned_data={
                "gender": MALE,
                "patient_log": self.patient_log(gender=MALE),
                "consent_ability": NO,
            },
            instance=SubjectScreeningMockModel(),
            model=SubjectScreeningMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn(
            "You may NOT screen this subject without their verbal consent",
            "|".join(cm.exception.messages),
        )

    def test_in_care_duration(self):
        form_validator = self.get_form_validator_cls()(
            cleaned_data={
                "report_datetime": get_utcnow(),
                "gender": MALE,
                "patient_log": self.patient_log(gender=MALE),
                "consent_ability": YES,
                "in_care_6m": YES,
                "in_care_duration": "1d",
                "hiv_dx": NO,
                "dm_dx": YES,
                "htn_dx": NO,
            },
            instance=SubjectScreeningMockModel(),
            model=SubjectScreeningMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn(
            "Expected at least 6m from the report date",
            "|".join(cm.exception.messages),
        )

    def test_in_care_duration_format(self):
        form_validator = self.get_form_validator_cls()(
            cleaned_data={
                "report_datetime": get_utcnow(),
                "gender": MALE,
                "patient_log": self.patient_log(gender=MALE),
                "consent_ability": YES,
                "in_care_6m": YES,
                "in_care_duration": "ERIK",
                "hiv_dx": NO,
                "dm_dx": YES,
                "htn_dx": NO,
            },
            instance=SubjectScreeningMockModel(),
            model=SubjectScreeningMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn(
            "Invalid format",
            "|".join(cm.exception.messages),
        )

    def test_no_conditions_in_patient_log(self):
        form_validator = self.get_form_validator_cls()(
            cleaned_data={
                "gender": MALE,
                "patient_log": self.patient_log(gender=MALE, conditions=[]),
                "consent_ability": YES,
                "in_care_6m": YES,
                "in_care_duration": "5y",
                "hiv_dx": YES,
            },
            instance=SubjectScreeningMockModel(),
            model=SubjectScreeningMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn(
            "No conditions (HIV/DM/HTN) have been indicated for this patient",
            "|".join(cm.exception.messages),
        )

    def test_conditions_not_matching_patient_log(self):
        form_validator = self.get_form_validator_cls()(
            cleaned_data={
                "gender": MALE,
                "patient_log": self.patient_log(gender=MALE),
                "consent_ability": YES,
                "in_care_6m": YES,
                "in_care_duration": "5y",
                "hiv_dx": YES,
            },
            instance=SubjectScreeningMockModel(),
            model=SubjectScreeningMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn(
            "Invalid. HIV was not indicated as a condition on the Patient Log",
            "|".join(cm.exception.messages),
        )

    def test_conditions_matching_patient_log(self):
        form_validator = self.get_form_validator_cls()(
            cleaned_data={
                "gender": MALE,
                "patient_log": self.patient_log(gender=MALE),
                "consent_ability": YES,
                "in_care_6m": YES,
                "in_care_duration": "5y",
                "hiv_dx": NO,
                "dm_dx": YES,
                "htn_dx": NO,
            },
            instance=SubjectScreeningMockModel(),
            model=SubjectScreeningMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError:
            self.fail("ValidationError unexpectedly raised")

    def test_conditions_durations(self):
        cleaned_data = {
            "report_datetime": get_utcnow(),
            "gender": MALE,
            "patient_log": self.patient_log(gender=MALE),
            "consent_ability": YES,
            "in_care_6m": YES,
            "in_care_duration": "5y",
            "hiv_dx": NO,
            "dm_dx": YES,
            "htn_dx": NO,
            "dm_dx_ago": "5m",
        }
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=SubjectScreeningMockModel(),
            model=SubjectScreeningMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn(
            "Expected at least 6m from the report date",
            "|".join(cm.exception.messages),
        )

        cleaned_data.update({"dm_dx_ago": "ERIK"})
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=SubjectScreeningMockModel(),
            model=SubjectScreeningMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn(
            "Invalid format",
            "|".join(cm.exception.messages),
        )

        cleaned_data.update({"dm_dx_ago": "6m"})
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=SubjectScreeningMockModel(),
            model=SubjectScreeningMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError:
            self.fail("ValidationError unexpectedly raised")

    def test_conditions_not_matching_patient_log2(self):
        form_validator = self.get_form_validator_cls()(
            cleaned_data={
                "gender": MALE,
                "patient_log": self.patient_log(gender=MALE),
                "consent_ability": YES,
                "in_care_6m": YES,
                "in_care_duration": "5y",
                "dm_dx": NO,
            },
            instance=SubjectScreeningMockModel(),
            model=SubjectScreeningMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn(
            "Invalid. DM was indicated as a condition on the Patient Log",
            "|".join(cm.exception.messages),
        )

    def test_eligibility(self):
        pass

    def test_reasons_unsuitable_required_if_unsuitable_for_study_yes(self):
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "unsuitable_for_study": YES,
                "reasons_unsuitable": "",
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=SubjectScreeningMockModel(),
            model=SubjectScreeningMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("reasons_unsuitable", cm.exception.error_dict)
        self.assertIn(
            "This field is required.",
            str(cm.exception.error_dict.get("reasons_unsuitable")),
        )

    def test_reasons_unsuitable_not_required_if_unsuitable_for_study_not_yes(self):
        with self.subTest():
            cleaned_data = self.get_cleaned_data()
            cleaned_data.update(
                {
                    "unsuitable_for_study": NO,
                    "reasons_unsuitable": "Some reason ....",
                }
            )
            form_validator = self.get_form_validator_cls()(
                cleaned_data=cleaned_data,
                instance=SubjectScreeningMockModel(),
                model=SubjectScreeningMockModel,
            )
            with self.assertRaises(forms.ValidationError) as cm:
                form_validator.validate()
            self.assertIn("reasons_unsuitable", cm.exception.error_dict)
            self.assertIn(
                "This field is not required.",
                str(cm.exception.error_dict.get("reasons_unsuitable")),
            )

    def test_unsuitable_agreed_applicable_if_unsuitable_for_study_yes(self):
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "unsuitable_for_study": YES,
                "reasons_unsuitable": "Some reason",
                "unsuitable_agreed": NOT_APPLICABLE,
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=SubjectScreeningMockModel(),
            model=SubjectScreeningMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("unsuitable_agreed", cm.exception.error_dict)
        self.assertIn(
            "This field is applicable.",
            str(cm.exception.error_dict.get("unsuitable_agreed")),
        )

    def test_unsuitable_agreed_not_applicable_if_unsuitable_for_study_no(self):
        for agreed_answ in [YES, NO]:
            with self.subTest(unsuitable_agreed=agreed_answ):
                cleaned_data = self.get_cleaned_data()
                cleaned_data.update(
                    {
                        "unsuitable_for_study": NO,
                        "reasons_unsuitable": "",
                        "unsuitable_agreed": agreed_answ,
                    }
                )
                form_validator = self.get_form_validator_cls()(
                    cleaned_data=cleaned_data,
                    instance=SubjectScreeningMockModel(),
                    model=SubjectScreeningMockModel,
                )
                with self.assertRaises(forms.ValidationError) as cm:
                    form_validator.validate()
                self.assertIn("unsuitable_agreed", cm.exception.error_dict)
                self.assertIn(
                    "This field is not applicable.",
                    str(cm.exception.error_dict.get("unsuitable_agreed")),
                )

    def test_unsuitable_agreed_no_raises_if_unsuitable_for_study_yes(self):
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "unsuitable_for_study": YES,
                "reasons_unsuitable": "Reason unsuitable",
                "unsuitable_agreed": NO,
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=SubjectScreeningMockModel(),
            model=SubjectScreeningMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("unsuitable_agreed", cm.exception.error_dict)
        self.assertIn(
            (
                "The study coordinator MUST agree with your assessment. "
                "Please discuss before continuing."
            ),
            str(cm.exception.error_dict.get("unsuitable_agreed")),
        )

    def test_unsuitable_agreed_yes_with_unsuitable_for_study_yes_ok(self):
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "unsuitable_for_study": YES,
                "reasons_unsuitable": "Reason unsuitable",
                "unsuitable_agreed": YES,
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=SubjectScreeningMockModel(),
            model=SubjectScreeningMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_first_health_talk_reponse_from_patientlog(self):
        cleaned_data = self.get_cleaned_data()
        cleaned_data["patient_log"].first_health_talk = TBD
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=SubjectScreeningMockModel(),
            model=SubjectScreeningMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("Has patient attended the first health talk", str(cm.exception))

    def test_second_health_talk_reponse_from_patientlog(self):
        cleaned_data = self.get_cleaned_data()
        cleaned_data["patient_log"].first_health_talk = NO
        cleaned_data["patient_log"].second_health_talk = TBD
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=SubjectScreeningMockModel(),
            model=SubjectScreeningMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("Has patient attended the second health talk", str(cm.exception))

    def test_health_talk_reponse_from_patientlog(self):
        cleaned_data = self.get_cleaned_data()
        cleaned_data["patient_log"].first_health_talk = NO
        cleaned_data["patient_log"].second_health_talk = NO
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=SubjectScreeningMockModel(),
            model=SubjectScreeningMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")



from django import forms
from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from edc_adverse_event.form_validator_mixins import (
    RequiresDeathReportFormValidatorMixin,
)
from edc_consent.constants import CONSENT_WITHDRAWAL
from edc_constants.constants import CLINICAL_WITHDRAWAL, DEAD, INVESTIGATOR_DECISION
from edc_form_validators import INVALID_ERROR, FormValidator
from edc_ltfu.constants import LTFU
from edc_ltfu.modelform_mixins import RequiresLtfuFormValidatorMixin
from edc_offstudy.constants import COMPLETED_FOLLOWUP
from edc_offstudy.exceptions import OffstudyError
from edc_prn.modelform_mixins import PrnFormValidatorMixin
from edc_transfer.constants import TRANSFERRED
from edc_utils import formatted_date
from edc_visit_schedule.constants import MONTH12
from edc_visit_schedule.utils import off_all_schedules_or_raise


class EndOfStudyFormValidator(
    RequiresDeathReportFormValidatorMixin,
    RequiresLtfuFormValidatorMixin,
    PrnFormValidatorMixin,
    FormValidator,
):
    death_report_model = "intecomm_ae.deathreport"
    ltfu_model = None

    def clean(self):
        self.confirm_off_all_schedules()
        self.validate_offstudy_datetime_against_last_seen_date()

        self.validate_completed_followup()

        self.required_if(DEAD, field="offstudy_reason", field_required="death_date")
        self.validate_death_report_if_deceased()

        # TODO: should not be selected if before 12 months from baseline
        self.required_if(TRANSFERRED, field="offstudy_reason", field_required="transfer_date")
        self.validate_transfer()

        # TODO: should not be selected if before 12 months from baseline
        self.required_if(LTFU, field="offstudy_reason", field_required="ltfu_date")
        self.validate_ltfu()

        self.applicable_if(
            CLINICAL_WITHDRAWAL,
            field="offstudy_reason",
            field_applicable="clinical_withdrawal_reason",
        )

        self.validate_other_specify(
            other_stored_value=INVESTIGATOR_DECISION,
            field="clinical_withdrawal_reason",
            other_specify_field="clinical_withdrawal_investigator_decision",
        )

        self.validate_other_specify(
            field="clinical_withdrawal_reason",
            other_specify_field="clinical_withdrawal_reason_other",
        )

        self.required_if(
            CONSENT_WITHDRAWAL,
            field="offstudy_reason",
            field_required="consent_withdrawal_reason",
        )

    def validate_completed_followup(self):
        if (
            self.cleaned_data.get("offstudy_reason")
            and self.cleaned_data.get("offstudy_reason").name == COMPLETED_FOLLOWUP
        ):
            subject_visit_model_cls = django_apps.get_model("intecomm_subject.subjectvisit")
            try:
                subject_visit_model_cls.objects.get(
                    subject_identifier=self.subject_identifier,
                    visit_code=MONTH12,
                    visit_code_sequence=0,
                )
            except ObjectDoesNotExist:
                self.raise_validation_error(
                    {"offstudy_reason": "Invalid. 12 month visit has not been submitted."},
                    INVALID_ERROR,
                )

    def confirm_off_all_schedules(self):
        try:
            off_all_schedules_or_raise(
                subject_identifier=self.cleaned_data.get("subject_identifier"),
            )
        except OffstudyError as e:
            self.raise_validation_error(str(e), INVALID_ERROR)

    def validate_offstudy_datetime_against_last_seen_date(self):
        if self.cleaned_data.get("offstudy_datetime") and self.cleaned_data.get(
            "last_seen_date"
        ):
            if (
                self.cleaned_data.get("last_seen_date")
                > self.cleaned_data.get("offstudy_datetime").date()
            ):
                raise forms.ValidationError(
                    {"last_seen_date": "Invalid. May not be after termination date"}
                )

    def validate_transfer(self):
        if (
            self.cleaned_data.get("offstudy_reason")
            and self.cleaned_data.get("offstudy_reason").name == TRANSFERRED
        ):
            transfer_model_cls = django_apps.get_model("intecomm_prn.subjecttransfer")
            try:
                obj = transfer_model_cls.objects.get(
                    subject_visit__subject_identifier=self.subject_identifier,
                )
            except ObjectDoesNotExist:
                self.raise_validation_error(
                    {
                        "offstudy_reason": (
                            f"Invalid. {transfer_model_cls._meta.verbose_name} "
                            "has not been submitted."
                        )
                    },
                    INVALID_ERROR,
                )
            else:
                if obj.transfer_date.date() != self.cleaned_data.get("transfer_date"):
                    dt = formatted_date(obj.transfer_date)
                    self.raise_validation_error(
                        {
                            "transfer_date": (
                                "Invalid. Does not match date on "
                                f"{transfer_model_cls._meta.verbose_name}. Expected {dt}."
                            )
                        },
                        INVALID_ERROR,
                    )



from django import forms
from edc_constants.constants import OTHER, PENDING, YES
from edc_crf.crf_form_validator_mixins import CrfFormValidatorMixin
from edc_dx_review.medical_date import DxDate, MedicalDateError, RxDate
from edc_dx_review.utils import raise_if_clinical_review_does_not_exist
from edc_form_validators import FormValidator


class HivInitialReviewFormValidator(
    CrfFormValidatorMixin,
    FormValidator,
):
    def __init__(self, **kwargs):
        self.dx_date = None
        self.rx_init_date = None
        super().__init__(**kwargs)

    def clean(self):
        raise_if_clinical_review_does_not_exist(self.cleaned_data.get("subject_visit"))

        try:
            self.dx_date = DxDate(self.cleaned_data)
        except MedicalDateError as e:
            self.raise_validation_error(e.message_dict, e.code)

        self.applicable_if(YES, field="receives_care", field_applicable="clinic")

        self.required_if(OTHER, field="clinic", field_required="clinic_other")

        self.applicable_if(YES, field="receives_care", field_applicable="rx_init")

        self.validate_rx_init_date()

        self.required_if(YES, field="rx_init", field_required="has_vl")

        self.validate_viral_load()

        self.required_if(YES, field="rx_init", field_required="has_cd4")

        self.validate_cd4()

    def validate_rx_init_date(self):
        if self.cleaned_data.get("rx_init") == YES:
            try:
                self.rx_init_date = RxDate(self.cleaned_data, reference_date=self.dx_date)
            except MedicalDateError as e:
                self.raise_validation_error(e.message_dict, e.code)
        else:
            self.not_required_if_true(
                self.cleaned_data.get("rx_init_date"),
                field="rx_init_date",
            )
            self.not_required_if_true(
                self.cleaned_data.get("rx_init_ago"),
                field="rx_init_ago",
            )

    def validate_viral_load(self):
        self.required_if(YES, PENDING, field="has_vl", field_required="drawn_date")
        if self.cleaned_data.get("drawn_date") and self.dx_date:
            if self.cleaned_data.get("drawn_date") < self.dx_date:
                raise forms.ValidationError(
                    {"drawn_date": "Invalid. Cannot be before HIV diagnosis."}
                )
        self.required_if(YES, field="has_vl", field_required="vl")
        self.required_if(YES, field="has_vl", field_required="vl_quantifier")

    def validate_cd4(self):
        self.required_if(YES, field="has_cd4", field_required="cd4")
        self.required_if(YES, field="has_cd4", field_required="cd4_date")
        if self.cleaned_data.get("cd4_date") and self.dx_date:
            if self.cleaned_data.get("cd4_date") < self.dx_date:
                raise forms.ValidationError(
                    {"cd4_date": "Invalid. Cannot be before HIV diagnosis."}
                )



from __future__ import annotations

from decimal import Decimal
from typing import Tuple

from django.conf import settings
from django.db.models import QuerySet
from django.utils.html import format_html
from edc_constants.constants import DM, HIV, HTN, YES
from edc_utils.round_up import round_up
from edc_visit_schedule.constants import MONTH12


class PatientNotStableError(Exception):
    pass


class PatientNotScreenedError(Exception):
    pass


class PatientNotConsentedError(Exception):
    pass


class PatientGroupRatioError(Exception):
    pass


class PatientGroupSizeError(Exception):
    pass


class PatientGroupMakeupError(Exception):
    pass


def is_end_of_study(subject_visit):
    return subject_visit.visit_code == MONTH12 and subject_visit.visit_code_sequence == 0


def get_min_group_size() -> int:
    """Returns minimum number of patients needed to form a group"""
    return getattr(settings, "INTECOMM_MIN_GROUP_SIZE", 14)


def get_group_size_for_ratio() -> int:
    """Returns number of patients to use to determine the condition
     ratio.

    Additional group members can be of any condition.
    """
    return getattr(settings, "INTECOMM_MIN_GROUP_SIZE_FOR_RATIO", 9)


def verify_patient_group_ratio_raise(
    patients: iter, raise_on_outofrange: bool | None = None
) -> Tuple[int, int, Decimal, bool]:
    ncd = 0.0
    hiv = 0.0
    outofrange = False
    raise_on_outofrange = True if raise_on_outofrange is None else raise_on_outofrange
    for patient_log in patients:
        if patient_log.conditions.all().count() == 1:
            if (
                patient_log.conditions.filter(name__in=[HTN, DM])
                .exclude(name__in=[HIV])
                .exists()
            ):
                ncd += 1.0
            if (
                patient_log.conditions.filter(name__in=[HIV])
                .exclude(name__in=[HTN, DM])
                .exists()
            ):
                hiv += 1.0
    if not ncd or not hiv:
        ratio = 0.0
    else:
        ratio = ncd / hiv
    if not (2.0 <= ratio <= 2.8):
        outofrange = True
        if raise_on_outofrange:
            raise PatientGroupRatioError(
                f"Ratio NDC:HIV not met. Expected at least 2:1. Got {int(ncd)}:{int(hiv)}. "
            )
    ncd = int(ncd)
    hiv = int(hiv)
    return int(ncd), int(hiv), round_up(Decimal(str(ratio)), Decimal("1.00")), outofrange


def confirm_patient_group_size_or_raise(
    patients: QuerySet | None = None,
    bypass_group_size_min: bool | None = None,
    group_count_min: int | None = None,
) -> None:
    """Confirm at least 14 if complete or override."""
    group_count_min = group_count_min or get_min_group_size()
    if not patients:
        raise PatientGroupSizeError("Patient group has no patients.")
    elif not bypass_group_size_min and patients.count() < group_count_min:
        raise PatientGroupSizeError(
            f"Patient group must have at least {group_count_min} patients. "
            f"Got {patients.count()}."
        )


def confirm_patient_group_minimum_of_each_condition_or_raise(
    patients: QuerySet | None = None,
) -> None:
    """Confirm at least 2 of each condition

    A minimum of two people with each of HIV, diabetes and
    hypertension will be selected per group. The rest will be
    patients with a variable mixture of the three conditions
    """
    hiv_only = 0
    for patient in patients.all():
        hiv_only += patient.conditions.filter(name=HIV).exclude(name__in=[DM, HTN]).count()
        if hiv_only >= 2:
            break
    if hiv_only < 2:
        raise PatientGroupMakeupError(
            f"Patient group must have at least 2 HIV only patients. Got {hiv_only}."
        )

    ncd_only = 0
    for patient in patients.all():
        ncd_only += patient.conditions.filter(name__in=[DM, HTN]).exclude(name=HIV).count()
        if ncd_only >= 4:
            break
    if ncd_only < 4:
        raise PatientGroupMakeupError(
            f"Patient group must have at least 4 NCD only patients. Got {ncd_only}."
        )


def confirm_patient_group_ratio_or_raise(
    patients: QuerySet | None = None,
    bypass_group_ratio: bool | None = None,
):
    if not bypass_group_ratio:
        verify_patient_group_ratio_raise(patients)


def confirm_patients_stable_and_screened_and_consented_or_raise(
    patients: QuerySet | None = None,
):
    if not patients:
        raise PatientGroupSizeError("Patient group has no patients.")
    else:
        for patient_log in patients.all():
            link = format_html(
                f'<a href="{patient_log.get_changelist_url()}?'
                f'q={str(patient_log.id)}">{patient_log}</a>'
            )
            if patient_log.stable != YES:
                errmsg = format_html(
                    "Patient is not known to be stable and in-care. "
                    f"See patient log for {link}."
                )
                raise PatientNotStableError(errmsg)
            if not patient_log.screening_identifier:
                errmsg = format_html(f"Patient has not screened for eligibility. See {link}.")
                raise PatientNotScreenedError(errmsg)
            if not patient_log.subject_identifier:
                errmsg = format_html(f"Patient has not consented. See {link}.")
                raise PatientNotConsentedError(errmsg)



from django import forms
from edc_consent.form_validators import SubjectConsentFormValidatorMixin
from edc_form_validators import FormValidator


class SubjectConsentFormValidator(SubjectConsentFormValidatorMixin, FormValidator):
    def validate_identity(self) -> None:
        """Override to validate `identity_type` is a hospital
        number and `identity` matches the screening form.
        """
        if self.cleaned_data.get("identity_type") != "hospital_no":
            raise forms.ValidationError({"identity_type": "Expected 'hospital number'."})



from edc_constants.constants import NO, YES
from edc_crf.crf_form_validator import CrfFormValidator
from edc_dx import get_diagnosis_labels
from edc_dx_review.form_validator_mixins import ClinicalReviewBaselineFormValidatorMixin
from edc_screening.utils import get_subject_screening_model_cls

INVALID_DX = "INVALID_DX"
PROTOCOL_INCIDENT = "PROTOCOL_INCIDENT"


class ClinicalReviewBaselineFormValidator(
    ClinicalReviewBaselineFormValidatorMixin, CrfFormValidator
):
    def clean(self) -> None:
        protocol_incident = NO
        for cond, label in get_diagnosis_labels().items():
            if (
                self.dx(cond)
                and self.subject_screening_dx(cond)
                and self.dx_at_screening(cond)
            ):
                self.validate_dx_at_screening_or_raise(cond, label)
                if self.dx_at_screening(cond) == YES and self.dx(cond) == NO:
                    protocol_incident = YES
                elif self.dx_at_screening(cond) == NO and self.dx(cond) == YES:
                    protocol_incident = YES
        if (
            self.cleaned_data.get("protocol_incident")
            and self.cleaned_data.get("protocol_incident") != protocol_incident
        ):
            self.raise_validation_error(
                {"protocol_incident": f"Expected {protocol_incident}"}, PROTOCOL_INCIDENT
            )

    @property
    def subject_screening(self):
        return get_subject_screening_model_cls().objects.get(
            subject_identifier=self.subject_identifier
        )

    def subject_screening_dx(self, cond: str):
        return getattr(self.subject_screening, f"{cond}_dx", "")

    def dx(self, cond: str):
        return self.cleaned_data.get(f"{cond}_dx")

    def dx_at_screening(self, cond: str):
        return self.cleaned_data.get(f"{cond}_dx_at_screening")

    def validate_dx_at_screening_or_raise(self, cond, label):
        if self.subject_screening_dx(cond) != self.dx_at_screening(cond):
            if self.subject_screening_dx(cond) == YES:
                self.raise_validation_error(
                    {
                        f"{cond}_dx_at_screening": (
                            f"{label.title()} diagnosis was reported at screening"
                        )
                    },
                    INVALID_DX,
                )
            elif self.subject_screening_dx(cond) == NO:
                self.raise_validation_error(
                    {
                        f"{cond}_dx_at_screening": (
                            f"{label.title()} diagnosis was not reported at screening"
                        )
                    },
                    INVALID_DX,
                )

    def create_protocol_incident(self):
        pass



from edc_crf.crf_form_validator_mixins import CrfFormValidatorMixin
from edc_dx_review.utils import medications_exists_or_raise
from edc_form_validators import FormValidator

from .mixins import DrugRefillFormValidatorMixin


class DrugRefillDmFormValidator(
    DrugRefillFormValidatorMixin, CrfFormValidatorMixin, FormValidator
):
    def clean(self):
        medications_exists_or_raise(self.cleaned_data.get("subject_visit"))
        self.validate_rx_as_m2m()
        self.validate_modifications()



from __future__ import annotations

from unittest.mock import patch

from django import forms
from edc_constants.constants import YES

from intecomm_form_validators.subject import SocialHarmsFormValidator as Base

from ..mock_models import SocialHarmsMockModel
from ..test_case_mixin import TestCaseMixin


class SocialHarmsTests(TestCaseMixin):
    @staticmethod
    def get_form_validator_cls():
        class SocialHarmsFormValidator(Base):
            pass

        return SocialHarmsFormValidator

    @patch("edc_dx_review.utils.raise_if_clinical_review_does_not_exist")
    def test_ok(self, mock_func):
        social_harms = SocialHarmsMockModel()
        cleaned_data = {
            "subject_visit": None,
            "partner": YES,
            "partner_disclosure": None,
        }
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=social_harms,
            model=SocialHarmsMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("partner_disclosure", cm.exception.error_dict)

        cleaned_data.update(
            {
                "partner_disclosure": "blah",
                "partner_impact": YES,
                "partner_impact_severity": None,
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=social_harms,
            model=SocialHarmsMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("partner_impact_severity", cm.exception.error_dict)

        cleaned_data.update(
            {
                "partner_impact_severity": "blah",
                "healthcare_impact": YES,
                "healthcare_impact_severity": None,
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=social_harms,
            model=SocialHarmsMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("healthcare_impact_severity", cm.exception.error_dict)

        cleaned_data.update(
            {
                "partner_impact_severity": "blah",
                "healthcare_impact": YES,
                "healthcare_impact_severity": "blah",
                "other_impact": YES,
                "other_impact_description": None,
                "other_impact_severity": None,
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=social_harms,
            model=SocialHarmsMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("other_impact_description", cm.exception.error_dict)

        cleaned_data.update(
            {
                "partner_impact_severity": "blah",
                "healthcare_impact": YES,
                "healthcare_impact_severity": "blah",
                "other_impact": YES,
                "other_impact_description": "blah",
                "other_impact_severity": None,
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=social_harms,
            model=SocialHarmsMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("other_impact_severity", cm.exception.error_dict)

        cleaned_data.update(
            {
                "partner_impact_severity": "blah",
                "healthcare_impact": YES,
                "healthcare_impact_severity": "blah",
                "other_impact": YES,
                "other_impact_description": "blah",
                "other_impact_severity": "blah",
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=social_harms,
            model=SocialHarmsMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError:
            self.fail("ValidationError unexpectedly raised")



from edc_crf.crf_form_validator_mixins import CrfFormValidatorMixin
from edc_dx_review.utils import medications_exists_or_raise
from edc_form_validators import FormValidator
from edc_rx.utils import TotalDaysMismatch, validate_total_days

from .mixins import DrugRefillFormValidatorMixin

TOTAL_DAYS_MISMATCH_ERROR = "TOTAL_DAYS_MISMATCH"


class DrugRefillHivFormValidator(
    DrugRefillFormValidatorMixin, CrfFormValidatorMixin, FormValidator
):
    def clean(self):
        medications_exists_or_raise(self.cleaned_data.get("subject_visit"))
        self.validate_rx_as_fk()
        self.validate_modifications()

        try:
            validate_total_days(self, return_in_days=self.cleaned_data.get("return_in_days"))
        except TotalDaysMismatch as e:
            self.raise_validation_error(
                {
                    "clinic_days": str(e),
                    "club_days": str(e),
                    "purchased_days": str(e),
                },
                TOTAL_DAYS_MISMATCH_ERROR,
            )



from django import forms
from django.apps import apps as django_apps
from edc_constants.constants import (
    FREE_OF_CHARGE,
    INSURANCE,
    OTHER,
    OWN_CASH,
    PATIENT_CLUB,
    RELATIVE,
    YES,
)
from edc_crf.crf_form_validator_mixins import CrfFormValidatorMixin
from edc_dx.form_validators import DiagnosisFormValidatorMixin
from edc_dx_review.utils import raise_if_clinical_review_does_not_exist
from edc_form_validators import FormValidator


class HealthEconomicsFormValidator(
    DiagnosisFormValidatorMixin,
    CrfFormValidatorMixin,
    FormValidator,
):
    drug_pay_sources_model = "intecomm_lists.DrugPaySources"

    def clean(self):
        raise_if_clinical_review_does_not_exist(self.cleaned_data.get("subject_visit"))

        self.clean_education()

        self.clean_recv_drugs_by_duration("month")

        self.clean_recv_drugs_by_duration("today")

        self.required_if(YES, field="health_insurance", field_required="health_insurance_cost")

        self.required_if(YES, field="patient_club", field_required="patient_club_cost")

    @property
    def drug_pay_sources_model_cls(self):
        return django_apps.get_model(self.drug_pay_sources_model)

    def clean_education(self):
        cond = (
            self.cleaned_data.get("education_in_years") is not None
            and self.cleaned_data.get("education_in_years") > 0
        )

        if cond and self.cleaned_data.get("education_in_years") > self.age_in_years:
            raise forms.ValidationError(
                {
                    "education_in_years": (
                        "Cannot exceed subject's age. "
                        f"Got subject is {self.age_in_years} years old."
                    )
                }
            )

        self.required_if_true(cond, field_required="education_certificate")

        self.applicable_if_true(cond, field_applicable="primary_school")
        self.required_if(
            YES,
            field="primary_school",
            field_required="primary_school_in_years",
            field_required_evaluate_as_int=True,
        )
        self.applicable_if_true(cond, field_applicable="secondary_school")
        self.required_if(
            YES,
            field="secondary_school",
            field_required="secondary_school_in_years",
            field_required_evaluate_as_int=True,
        )
        self.applicable_if_true(cond, field_applicable="higher_education")
        self.required_if(
            YES,
            field="higher_education",
            field_required="higher_education_in_years",
            field_required_evaluate_as_int=True,
        )

    def clean_recv_drugs_by_duration(self, duration):
        conditions = [
            ("dm", "diabetes"),
            ("htn", "hypertension"),
            ("hiv", "HIV"),
            ("other", None),
        ]
        diagnoses = self.get_diagnoses()
        for cond, label in conditions:
            if cond == "other":
                self.applicable_if(
                    YES,
                    field=f"received_rx_{duration}",
                    field_applicable=f"rx_{cond}_{duration}",
                )
            else:
                if self.cleaned_data.get(f"received_rx_{duration}") == YES:
                    self.applicable_if_diagnosed(
                        diagnoses=diagnoses,
                        prefix=f"{cond}_dx",
                        field_applicable=f"rx_{cond}_{duration}",
                        label=label,
                    )
                else:
                    self.applicable_if(
                        YES,
                        field=f"received_rx_{duration}",
                        field_applicable=f"rx_{cond}_{duration}",
                    )

            self.m2m_required_if(
                response=YES,
                field=f"rx_{cond}_{duration}",
                m2m_field=f"rx_{cond}_paid_{duration}",
            )
            self.m2m_single_selection_if(
                FREE_OF_CHARGE, m2m_field=f"rx_{cond}_paid_{duration}"
            )
            self.m2m_other_specify(
                OTHER,
                m2m_field=f"rx_{cond}_paid_{duration}",
                field_other=f"rx_{cond}_paid_{duration}_other",
            )
            self.m2m_other_specify(
                *[
                    obj.name
                    for obj in (
                        # NOTE: Must align with choices in intecomm_lists.models.DrugPaySources
                        (OWN_CASH, INSURANCE, PATIENT_CLUB, RELATIVE, FREE_OF_CHARGE, OTHER)
                    )
                    if obj.name != FREE_OF_CHARGE
                ],
                m2m_field=f"rx_{cond}_paid_{duration}",
                field_other=f"rx_{cond}_cost_{duration}",
                field_other_evaluate_as_int=True,
            )



from django import forms
from edc_constants.constants import OTHER, YES
from edc_form_validators import FormValidator
from edc_visit_schedule.utils import is_baseline


class DrugRefillFormValidatorMixin(FormValidator):

    """For example:
    def clean(self):
        medications_exists_or_raise(self.cleaned_data.get("subject_visit"))
        self.validate_rx_as_m2m()
        self.validate_modifications()

    - or -

    def clean(self):
        medications_exists_or_raise(self.cleaned_data.get("subject_visit"))
        self.validate_rx_as_fk()
        self.validate_modifications()
    """

    def validate_rx_as_m2m(self):
        self.m2m_other_specify(OTHER, m2m_field="rx", field_other="rx_other")

    def validate_rx_as_fk(self):
        self.validate_other_specify(field="rx", other_specify_field="rx_other")

    def validate_modifications(self):
        if (
            self.cleaned_data.get("subject_visit")
            and is_baseline(self.cleaned_data.get("subject_visit"))
            and self.cleaned_data.get("rx_modified") == YES
        ):
            raise forms.ValidationError({"rx_modified": "Expected `No` at baseline."})
        self.m2m_required_if(YES, field="rx_modified", m2m_field="modifications")
        self.m2m_other_specify(
            OTHER, m2m_field="modifications", field_other="modifications_other"
        )
        self.m2m_required_if(YES, field="rx_modified", m2m_field="modification_reasons")
        self.m2m_other_specify(
            OTHER,
            m2m_field="modification_reasons",
            field_other="modification_reasons_other",
        )


class BPFormValidatorMixin:
    def validate_bp_reading(self, sys_field, dia_field):
        if self.cleaned_data.get(sys_field) and self.cleaned_data.get(dia_field):
            if self.cleaned_data.get(sys_field) < self.cleaned_data.get(dia_field):
                raise forms.ValidationError(
                    {dia_field: "Systolic must be greater than diastolic."}
                )



from edc_crf.crf_form_validator_mixins import CrfFormValidatorMixin
from edc_dx_review.constants import DRUGS, INSULIN
from edc_dx_review.medical_date import DxDate, MedicalDateError, RxDate
from edc_dx_review.utils import raise_if_clinical_review_does_not_exist
from edc_form_validators import INVALID_ERROR
from edc_form_validators.form_validator import FormValidator


class HtnInitialReviewFormValidator(
    CrfFormValidatorMixin,
    FormValidator,
):
    def clean(self):
        raise_if_clinical_review_does_not_exist(self.cleaned_data.get("subject_visit"))

        try:
            dx_date = DxDate(self.cleaned_data)
        except MedicalDateError as e:
            self.raise_validation_error(e.message_dict, e.code)

        if DRUGS in self.get_m2m_selected("managed_by") or INSULIN in self.get_m2m_selected(
            "managed_by"
        ):
            try:
                RxDate(self.cleaned_data, reference_date=dx_date)
            except MedicalDateError as e:
                self.raise_validation_error(e.message_dict, e.code)
        else:
            if self.cleaned_data.get("rx_init_date"):
                self.raise_validation_error(
                    {"rx_init_date": "This field is not required"}, INVALID_ERROR
                )
            if self.cleaned_data.get("rx_init_ago"):
                self.raise_validation_error(
                    {"rx_init_ago": "This field is not required"}, INVALID_ERROR
                )

        self.m2m_other_specify(m2m_field="managed_by", field_other="managed_by_other")

        try:
            RxDate(self.cleaned_data, reference_date=dx_date)
        except MedicalDateError as e:
            self.raise_validation_error(e.message_dict, e.code)



from dateutil.relativedelta import relativedelta
from edc_crf.crf_form_validator import CrfFormValidator
from edc_dx_review.utils import raise_if_clinical_review_does_not_exist
from edc_form_validators import INVALID_ERROR


class NextAppointmentFormValidator(CrfFormValidator):
    def __init__(self, **kwargs):
        self._integrated_clinic_days = None
        super().__init__(**kwargs)

    def clean(self):
        raise_if_clinical_review_does_not_exist(self.cleaned_data.get("subject_visit"))
        self.date_not_before(
            "report_datetime",
            "appt_date",
            msg="Cannot be before the report date",
            convert_to_date=True,
        )
        self.date_not_equal(
            "report_datetime",
            "appt_date",
            msg="Cannot be the same as the report date",
            convert_to_date=True,
        )

        if self.cleaned_data.get("report_datetime") and self.cleaned_data.get("appt_date"):
            report_date = self.cleaned_data.get("report_datetime").date()
            if self.cleaned_data.get("appt_date") > report_date + relativedelta(months=6):
                raise self.raise_validation_error(
                    {"appt_date": "Cannot be more than 6 months from report date"},
                    INVALID_ERROR,
                )
        self.validate_date_is_on_clinic_day()

    @property
    def integrated_clinic_days(self) -> list[int]:
        if not self._integrated_clinic_days:
            if self.cleaned_data.get("health_facility"):
                self._integrated_clinic_days = self.cleaned_data.get(
                    "health_facility"
                ).clinic_days
            # if not self._integrated_clinic_days:
            #     site = self.cleaned_data.get("subject_visit").site
            #     if s_obj := [s for s in all_sites.get("uganda") if s.site_id == site.id]:
            #         self._integrated_clinic_days = s_obj[0].integrated_clinic_days
            #     else:
            #         self._integrated_clinic_days = [
            #             s for s in all_sites.get("tanzania") if s.site_id == site.id
            #         ][0].integrated_clinic_days
        return self._integrated_clinic_days

    def validate_date_is_on_clinic_day(self):
        if appt_date := self.cleaned_data.get("appt_date"):
            if appt_date.isoweekday() > 5:
                day = "Saturday" if appt_date.isoweekday() == 6 else "Sunday"
                raise self.raise_validation_error(
                    {"appt_date": f"Expected Mon-Fri. Got {day}"},
                    INVALID_ERROR,
                )
        if appt_date and self.cleaned_data.get("subject_visit").site:
            if (
                self.integrated_clinic_days
                and appt_date.isoweekday() not in self.integrated_clinic_days
            ):
                dct = dict(zip([1, 2, 3, 4, 5, 6], ["M", "T", "W", "Th", "F", "Sa"]))
                expected_days = [v for k, v in dct.items() if k in self.integrated_clinic_days]
                raise self.raise_validation_error(
                    {
                        "appt_date": (
                            "Not an integrated clinic day. "
                            f"Expected {''.join(expected_days)}. "
                            f"Got {appt_date.strftime('%A')}"
                        )
                    },
                    INVALID_ERROR,
                )



from .clinical_review_baseline_form_validator import ClinicalReviewBaselineFormValidator
from .complications_baseline_form_validator import ComplicationsBaselineFormValidator
from .dm_initial_review_form_validator import DmInitialReviewFormValidator
from .dm_medication_adherence_form_validator import DmMedicationAdherenceFormValidator
from .dm_review_form_validator import DmReviewFormValidator
from .drug_refill_dm_form_validator import DrugRefillDmFormValidator
from .drug_refill_hiv_form_validator import DrugRefillHivFormValidator
from .drug_refill_htn_form_validator import DrugRefillHtnFormValidator
from .health_economics_form_validator import HealthEconomicsFormValidator
from .hiv_initial_review_form_validator import HivInitialReviewFormValidator
from .hiv_medication_adherence_form_validator import HivMedicationAdherenceFormValidator
from .hiv_review_form_validator import HivReviewFormValidator
from .htn_initial_review_form_validator import HtnInitialReviewFormValidator
from .htn_medication_adherence_form_validator import HtnMedicationAdherenceFormValidator
from .htn_review_form_validator import HtnReviewFormValidator
from .medications_form_validator import MedicationsFormValidator
from .next_appointment_form_validator import NextAppointmentFormValidator
from .other_baseline_data_form_validator import OtherBaselineDataFormValidator
from .social_harms_form_validator import SocialHarmsFormValidator
from .vitals_form_validator import VitalsFormValidator



from __future__ import annotations

from unittest.mock import patch

from dateutil.relativedelta import relativedelta
from dateutil.utils import today
from django import forms
from django_mock_queries.query import MockSet
from edc_constants.constants import NO, OTHER, YES
from edc_dx_review.constants import DIET_LIFESTYLE, DRUGS, INSULIN
from edc_reportable import MILLIMOLES_PER_LITER
from edc_utils import get_utcnow

from intecomm_form_validators.subject import DmInitialReviewFormValidator as DmBase

from ..mock_models import (
    AppointmentMockModel,
    DmInitialReviewMockModel,
    DmTreatmentsMockModel,
    SubjectVisitMockModel,
)
from ..test_case_mixin import TestCaseMixin


class InitialReviewTests(TestCaseMixin):
    @staticmethod
    def get_form_validator_cls():
        class DmInitialReviewFormValidator(DmBase):
            pass

        return DmInitialReviewFormValidator

    @patch("edc_dx_review.utils.raise_if_clinical_review_does_not_exist")
    def test_cannot_enter_ago_and_exact_date(self, mock_func):
        appointment = AppointmentMockModel()
        subject_visit = SubjectVisitMockModel(appointment)
        dm_initial_review = DmInitialReviewMockModel()
        cleaned_data = {
            "subject_visit": subject_visit,
            "report_datetime": get_utcnow(),
            "dx_ago": "2y",
            "dx_date": today(),
        }
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=dm_initial_review,
            model=DmInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("dx_ago", cm.exception.error_dict)

    @patch("edc_dx_review.utils.raise_if_clinical_review_does_not_exist")
    def test_if_managed_by_drugs_required_rx_init_ago(self, mock_func):
        appointment = AppointmentMockModel()
        subject_visit = SubjectVisitMockModel(appointment)
        dm_initial_review = DmInitialReviewMockModel()
        for managed_by in [DRUGS, INSULIN]:
            with self.subTest(managed_by=managed_by):
                cleaned_data = {
                    "subject_visit": subject_visit,
                    "report_datetime": get_utcnow(),
                    "dx_ago": "2y",
                    "managed_by": MockSet(DmTreatmentsMockModel(name=managed_by)).filter(
                        name=managed_by
                    ),
                }
                form_validator = self.get_form_validator_cls()(
                    cleaned_data=cleaned_data,
                    instance=dm_initial_review,
                    model=DmInitialReviewMockModel,
                )
                with self.assertRaises(forms.ValidationError) as cm:
                    form_validator.validate()
                self.assertIn("Complete the treatment date", str(cm.exception))

    @patch("edc_dx_review.utils.raise_if_clinical_review_does_not_exist")
    def test_if_managed_by_lifestyle(self, mock_func):
        appointment = AppointmentMockModel()
        subject_visit = SubjectVisitMockModel(appointment)
        dm_initial_review = DmInitialReviewMockModel()
        cleaned_data = {
            "subject_visit": subject_visit,
            "report_datetime": get_utcnow(),
            "dx_ago": "2y",
            "managed_by": MockSet(DmTreatmentsMockModel(name=DIET_LIFESTYLE)).filter(
                name=DIET_LIFESTYLE
            ),
            "rx_init_ago": "blah",
        }
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=dm_initial_review,
            model=DmInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("rx_init_ago", cm.exception.error_dict)
        self.assertIn("This field is not required", str(cm.exception))

    @patch("edc_dx_review.utils.raise_if_clinical_review_does_not_exist")
    def test_if_managed_by_other(self, mock_func):
        appointment = AppointmentMockModel()
        subject_visit = SubjectVisitMockModel(appointment)
        dm_initial_review = DmInitialReviewMockModel()
        cleaned_data = {
            "subject_visit": subject_visit,
            "report_datetime": get_utcnow(),
            "dx_ago": "2y",
            "managed_by": MockSet(DmTreatmentsMockModel(name=OTHER)).filter(name=OTHER),
            "rx_init_ago": None,
            "managed_by_other": None,
        }
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=dm_initial_review,
            model=DmInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("managed_by_other", cm.exception.error_dict)

    @patch("edc_dx_review.utils.raise_if_clinical_review_does_not_exist")
    def test_if_managed_by_drug_med_started_after_dx(self, mock_func):
        appointment = AppointmentMockModel()
        subject_visit = SubjectVisitMockModel(appointment)
        dm_initial_review = DmInitialReviewMockModel()
        cleaned_data = {
            "subject_visit": subject_visit,
            "report_datetime": get_utcnow(),
            "dx_ago": "2y",
            "managed_by": MockSet(DmTreatmentsMockModel(name=DRUGS)).filter(name=DRUGS),
            "rx_init_ago": "3y",
        }
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=dm_initial_review,
            model=DmInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("Treatment date must be on or after", str(cm.exception))

        cleaned_data.update(rx_init_ago="2y")
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=dm_initial_review,
            model=DmInitialReviewMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError:
            self.fail("ValidationError unexpectedly raised")

        cleaned_data.update(rx_init_ago="1y")
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=dm_initial_review,
            model=DmInitialReviewMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError:
            self.fail("ValidationError unexpectedly raised")

    @patch("edc_dx_review.utils.raise_if_clinical_review_does_not_exist")
    def test_glucose_not_tested_glucose_fasting_not_applicable(self, mock_func):
        appointment = AppointmentMockModel()
        subject_visit = SubjectVisitMockModel(appointment)
        dm_initial_review = DmInitialReviewMockModel()
        cleaned_data = {
            "subject_visit": subject_visit,
            "report_datetime": get_utcnow(),
            "dx_ago": "2y",
            "managed_by": MockSet(DmTreatmentsMockModel(name=DRUGS)).filter(name=DRUGS),
            "rx_init_ago": "2y",
            "glucose_performed": NO,
            "glucose_fasting": YES,
        }
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=dm_initial_review,
            model=DmInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("glucose_fasting", cm.exception.error_dict)
        self.assertIn("not applicable", str(cm.exception))

    @patch("edc_dx_review.utils.raise_if_clinical_review_does_not_exist")
    def test_glucose_tested_requires_fasting_duration(self, mock_func):
        appointment = AppointmentMockModel()
        subject_visit = SubjectVisitMockModel(appointment)
        dm_initial_review = DmInitialReviewMockModel()
        cleaned_data = {
            "subject_visit": subject_visit,
            "report_datetime": get_utcnow(),
            "dx_ago": "2y",
            "managed_by": MockSet(DmTreatmentsMockModel(name=DRUGS)).filter(name=DRUGS),
            "rx_init_ago": "2y",
            "glucose_performed": YES,
            "glucose_fasting": YES,
            "glucose_fasting_duration_str": None,
            "glucose_date": None,
        }
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=dm_initial_review,
            model=DmInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("glucose_fasting_duration_str", cm.exception.error_dict)

    @patch("edc_dx_review.utils.raise_if_clinical_review_does_not_exist")
    def test_glucose_tested_requires_date(self, mock_func):
        appointment = AppointmentMockModel()
        subject_visit = SubjectVisitMockModel(appointment)
        dm_initial_review = DmInitialReviewMockModel()
        cleaned_data = {
            "subject_visit": subject_visit,
            "report_datetime": get_utcnow(),
            "dx_ago": "2y",
            "managed_by": MockSet(DmTreatmentsMockModel(name=DRUGS)).filter(name=DRUGS),
            "rx_init_ago": "2y",
            "glucose_performed": YES,
            "glucose_fasting": YES,
            "glucose_fasting_duration_str": "8h",
            "glucose_date": None,
        }
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=dm_initial_review,
            model=DmInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("glucose_date", cm.exception.error_dict)

        cleaned_data = {
            "subject_visit": subject_visit,
            "report_datetime": get_utcnow(),
            "dx_ago": "2y",
            "managed_by": MockSet(DmTreatmentsMockModel(name=DRUGS)).filter(name=DRUGS),
            "rx_init_ago": "2y",
            "glucose_fasting": YES,
            "glucose_fasting_duration_str": "8h",
            "glucose_performed": YES,
            "glucose_value": 8.3,
            "glucose_quantifier": "=",
            "glucose_units": MILLIMOLES_PER_LITER,
        }

        for rdelta in [
            relativedelta(months=0),
        ]:
            cleaned_data.update(glucose_date=get_utcnow() + rdelta)
            with self.subTest(rdelta=rdelta):
                form_validator = self.get_form_validator_cls()(
                    cleaned_data=cleaned_data,
                    instance=dm_initial_review,
                    model=DmInitialReviewMockModel,
                )
                try:
                    form_validator.validate()
                except forms.ValidationError:
                    self.fail("ValidationError unexpectedly raised")

        for rdelta in [
            relativedelta(years=-1),
            relativedelta(months=-7),
            relativedelta(months=2),
            relativedelta(months=1),
        ]:
            cleaned_data.update(glucose_date=get_utcnow() + rdelta)
            with self.subTest(rdelta=rdelta):
                form_validator = self.get_form_validator_cls()(
                    cleaned_data=cleaned_data,
                    instance=dm_initial_review,
                    model=DmInitialReviewMockModel,
                )
                with self.assertRaises(forms.ValidationError) as cm:
                    form_validator.validate()
                self.assertIn("glucose_date", cm.exception.error_dict)



from edc_crf.crf_form_validator_mixins import CrfFormValidatorMixin
from edc_dx import get_diagnosis_labels
from edc_dx.form_validators import DiagnosisFormValidatorMixin
from edc_form_validators import FormValidator


class MedicationsFormValidator(
    DiagnosisFormValidatorMixin, CrfFormValidatorMixin, FormValidator
):
    def clean(self) -> None:
        for dx, label in get_diagnosis_labels().items():
            self.applicable_if_diagnosed(
                diagnoses=self.get_diagnoses(),
                prefix=dx,
                field_applicable=f"refill_{dx}",
                label=label,
            )



from edc_constants.constants import ESTIMATED, MEASURED, NO, YES
from edc_crf.crf_form_validator_mixins import CrfFormValidatorMixin
from edc_dx_review.utils import raise_if_clinical_review_does_not_exist
from edc_form_validators import INVALID_ERROR, FormValidator
from edc_visit_schedule.utils import is_baseline
from edc_vitals.form_validators import BloodPressureFormValidatorMixin

from intecomm_form_validators.utils import is_end_of_study


class VitalsFormValidator(
    BloodPressureFormValidatorMixin,
    CrfFormValidatorMixin,
    FormValidator,
):
    def clean(self):
        raise_if_clinical_review_does_not_exist(self.cleaned_data.get("subject_visit"))

        self.weight_required_at_baseline_and_eos()

        self.required_if(
            MEASURED, ESTIMATED, field="weight_determination", field_required="weight"
        )

        self.required_if_true(
            is_baseline(self.cleaned_data.get("subject_visit")),
            field_required="height",
            inverse=False,
            required_msg="(at this timepoint)",
        )

        self.required_if(YES, field="bp_one_taken", field_required="sys_blood_pressure_one")
        self.required_if(YES, field="bp_one_taken", field_required="dia_blood_pressure_one")
        self.required_if(NO, field="bp_one_taken", field_required="bp_one_not_taken_reason")
        self.raise_on_systolic_lt_diastolic_bp(
            sys_field="sys_blood_pressure_one",
            dia_field="dia_blood_pressure_one",
            **self.cleaned_data,
        )

        self.required_if(YES, field="bp_two_taken", field_required="sys_blood_pressure_two")
        self.required_if(YES, field="bp_two_taken", field_required="dia_blood_pressure_two")
        self.required_if(NO, field="bp_two_taken", field_required="bp_two_not_taken_reason")
        self.applicable_if(YES, field="bp_one_taken", field_applicable="bp_two_taken")
        self.raise_on_systolic_lt_diastolic_bp(
            sys_field="sys_blood_pressure_two",
            dia_field="dia_blood_pressure_two",
            **self.cleaned_data,
        )

        self.applicable_if(YES, field="bp_one_taken", field_applicable="severe_htn")
        opts = {
            "severe_htn": self.cleaned_data.get("severe_htn"),
            "sys_blood_pressure_one": self.cleaned_data.get("sys_blood_pressure_one"),
            "dia_blood_pressure_one": self.cleaned_data.get("dia_blood_pressure_one"),
            "sys_blood_pressure_two": (
                self.cleaned_data.get("sys_blood_pressure_two")
                if self.cleaned_data.get("sys_blood_pressure_two") is not None
                else self.cleaned_data.get("sys_blood_pressure_one")
            ),
            "dia_blood_pressure_two": (
                self.cleaned_data.get("dia_blood_pressure_two")
                if self.cleaned_data.get("dia_blood_pressure_two") is not None
                else self.cleaned_data.get("dia_blood_pressure_one")
            ),
        }
        self.raise_on_avg_blood_pressure_suggests_severe_htn(**opts)

    def weight_required_at_baseline_and_eos(self):
        if (
            self.cleaned_data.get("weight_determination")
            and self.cleaned_data.get("weight_determination") != MEASURED
            and (
                is_baseline(self.cleaned_data.get("subject_visit"))
                or is_end_of_study(self.cleaned_data.get("subject_visit"))
            )
        ):
            self.raise_validation_error(
                {"weight_determination": "Expected weight to be measured at this timepoint"},
                INVALID_ERROR,
            )



from edc_crf.crf_form_validator_mixins import CrfFormValidatorMixin
from edc_dx_review.utils import raise_if_clinical_review_does_not_exist
from edc_form_validators import FormValidator
from edc_glucose.form_validators import GlucoseFormValidatorMixin


class DmReviewFormValidator(
    GlucoseFormValidatorMixin,
    CrfFormValidatorMixin,
    FormValidator,
):
    prefix = "glucose"

    def clean(self):
        raise_if_clinical_review_does_not_exist(self.cleaned_data.get("subject_visit"))
        self.validate_glucose_test()



from edc_constants.constants import FORMER_SMOKER, SMOKER, YES
from edc_crf.crf_form_validator_mixins import CrfFormValidatorMixin
from edc_dx_review.utils import raise_if_clinical_review_does_not_exist
from edc_form_validators import FormValidator


class OtherBaselineDataFormValidator(CrfFormValidatorMixin, FormValidator):
    def clean(self):
        raise_if_clinical_review_does_not_exist(self.cleaned_data.get("subject_visit"))
        self.required_if(
            SMOKER, field="smoking_status", field_required="smoker_current_duration"
        )
        self.required_if(
            FORMER_SMOKER, field="smoking_status", field_required="smoker_quit_ago"
        )
        # self.estimated_date_from_ago("smoker_quit_ago")

        self.applicable_if(YES, field="alcohol", field_applicable="alcohol_consumption")

        self.required_if(
            YES, field="activity_work", field_required="activity_work_days_per_wk"
        )

        self.validate_other_specify(
            field="employment_status", other_specify_field="employment_status_other"
        )



from __future__ import annotations

from datetime import datetime
from unittest.mock import patch

from dateutil.relativedelta import relativedelta
from dateutil.utils import today
from django import forms
from django_mock_queries.query import MockSet
from edc_constants.constants import COMPLETE, OTHER
from edc_dx_review.constants import DIET_LIFESTYLE, DRUGS
from edc_utils import get_utcnow

from intecomm_form_validators.subject import HtnInitialReviewFormValidator as HtnBase

from ..mock_models import (
    AppointmentMockModel,
    HtnInitialReviewMockModel,
    HtnTreatmentsMockModel,
    SubjectVisitMockModel,
)
from ..test_case_mixin import TestCaseMixin


class HtnInitialReviewTests(TestCaseMixin):
    def setUp(self) -> None:
        super().setUp()
        raise_missing_clinical_review_patcher = patch(
            "edc_dx_review.utils.raise_if_clinical_review_does_not_exist"
        )
        self.addCleanup(raise_missing_clinical_review_patcher.stop)
        self.raise_missing_clinical_review = raise_missing_clinical_review_patcher.start()

        self.dx_date = get_utcnow() - relativedelta(days=60)
        self.rx_date = self.dx_date + relativedelta(days=3)

    @staticmethod
    def get_form_validator_cls():
        class HtnInitialReviewFormValidator(HtnBase):
            pass

        return HtnInitialReviewFormValidator

    def get_cleaned_data(self, report_datetime: datetime = None) -> dict:
        cleaned_data = dict(
            subject_visit=self.get_subject_visit(),
            report_datetime=report_datetime or get_utcnow(),
            dx_date=get_utcnow() - relativedelta(days=60),
            dx_ago="",
            managed_by=MockSet(HtnTreatmentsMockModel(name=DRUGS)).filter(name=DRUGS),
            managed_by_other="",
            rx_init_date=self.rx_date,
            rx_init_ago=None,
            crf_status=COMPLETE,
            crf_status_comments="",
        )
        return cleaned_data

    def test_cleaned_data_ok(self):
        htn_initial_review = HtnInitialReviewMockModel()
        cleaned_data = self.get_cleaned_data()
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=htn_initial_review,
            model=HtnInitialReviewMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_cannot_enter_dx_ago_and_exact_date(self):
        htn_initial_review = HtnInitialReviewMockModel()
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "dx_date": today(),
                "dx_ago": "2y",
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=htn_initial_review,
            model=HtnInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("dx_ago", cm.exception.error_dict)
        self.assertIn(
            "'Date conflict. Do not provide a response here if diagnosis date is available.'",
            str(cm.exception.error_dict.get("dx_ago")),
        )

    def test_dx_date_only_ok(self):
        htn_initial_review = HtnInitialReviewMockModel()
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "dx_date": self.dx_date,
                "dx_ago": "",
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=htn_initial_review,
            model=HtnInitialReviewMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_dx_ago_only_ok(self):
        htn_initial_review = HtnInitialReviewMockModel()
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "dx_date": None,
                "dx_ago": "1y",
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=htn_initial_review,
            model=HtnInitialReviewMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_dx_date_after_report_datetime_raises(self):
        htn_initial_review = HtnInitialReviewMockModel()
        report_datetime = get_utcnow() - relativedelta(days=30)
        cleaned_data = self.get_cleaned_data(report_datetime=report_datetime)
        cleaned_data.update(
            {
                "dx_date": report_datetime + relativedelta(days=1),
                "dx_ago": "",
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=htn_initial_review,
            model=HtnInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("dx_date", cm.exception.error_dict)
        self.assertIn(
            "Diagnosis date must be on or before ",
            str(cm.exception.error_dict.get("dx_date")),
        )

    def test_invalid_dx_ago_raises(self):
        htn_initial_review = HtnInitialReviewMockModel()
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "dx_date": None,
                "dx_ago": "1dx",
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=htn_initial_review,
            model=HtnInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("dx_ago", cm.exception.error_dict)
        self.assertIn(
            "Expected format ",
            str(cm.exception.error_dict.get("dx_ago")),
        )

    def test_if_managed_by_drugs_requires_rx_init_date(self):
        htn_initial_review = HtnInitialReviewMockModel()
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "managed_by": MockSet(HtnTreatmentsMockModel(name=DRUGS)).filter(name=DRUGS),
                "rx_init_date": None,
                "rx_init_ago": "",
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=htn_initial_review,
            model=HtnInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("Complete the treatment date", str(cm.exception))

        cleaned_data.update(
            {
                "managed_by": MockSet(HtnTreatmentsMockModel(name=DRUGS)).filter(name=DRUGS),
                "rx_init_date": None,
                "rx_init_ago": "14d",
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=htn_initial_review,
            model=HtnInitialReviewMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

        cleaned_data.update(
            {
                "managed_by": MockSet(HtnTreatmentsMockModel(name=DRUGS)).filter(name=DRUGS),
                "rx_init_date": get_utcnow() - relativedelta(days=7),
                "rx_init_ago": "",
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=htn_initial_review,
            model=HtnInitialReviewMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_rx_date_after_report_datetime_raises(self):
        htn_initial_review = HtnInitialReviewMockModel()
        report_datetime = get_utcnow() - relativedelta(days=30)
        cleaned_data = self.get_cleaned_data(report_datetime=report_datetime)
        cleaned_data.update(
            {
                "dx_date": report_datetime - relativedelta(days=7),
                "dx_ago": "",
                "rx_init_date": report_datetime + relativedelta(days=1),
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=htn_initial_review,
            model=HtnInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("rx_init_date", cm.exception.error_dict)
        self.assertIn(
            "Cannot be after report date",
            str(cm.exception.error_dict.get("rx_init_date")),
        )

    def test_invalid_rx_init_ago_raises(self):
        htn_initial_review = HtnInitialReviewMockModel()
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "managed_by": MockSet(HtnTreatmentsMockModel(name=DRUGS)).filter(name=DRUGS),
                "rx_init_date": None,
                "rx_init_ago": "1yxxx",
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=htn_initial_review,
            model=HtnInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("rx_init_ago", cm.exception.error_dict)
        self.assertIn(
            "Expected format ",
            str(cm.exception.error_dict.get("rx_init_ago")),
        )

    def test_rx_init_without_dx_date_raises(self):
        htn_initial_review = HtnInitialReviewMockModel()
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "dx_date": None,
                "dx_ago": "",
                "managed_by": MockSet(HtnTreatmentsMockModel(name=DRUGS)).filter(name=DRUGS),
                "rx_init_date": get_utcnow(),
                "rx_init_ago": "",
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=htn_initial_review,
            model=HtnInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("Complete the diagnosis date", str(cm.exception))

    def test_if_managed_by_lifestyle(self):
        appointment = AppointmentMockModel()
        subject_visit = SubjectVisitMockModel(appointment)
        htn_initial_review = HtnInitialReviewMockModel()
        cleaned_data = {
            "subject_visit": subject_visit,
            "report_datetime": get_utcnow(),
            "dx_ago": "2y",
            "managed_by": MockSet(HtnTreatmentsMockModel(name=DIET_LIFESTYLE)).filter(
                name=DIET_LIFESTYLE
            ),
            "rx_init_ago": "blah",
        }
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=htn_initial_review,
            model=HtnInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("rx_init_ago", cm.exception.error_dict)

    def test_if_managed_by_other(self):
        appointment = AppointmentMockModel()
        subject_visit = SubjectVisitMockModel(appointment)
        htn_initial_review = HtnInitialReviewMockModel()
        cleaned_data = {
            "subject_visit": subject_visit,
            "report_datetime": get_utcnow(),
            "dx_ago": "2y",
            "managed_by": MockSet(HtnTreatmentsMockModel(name=OTHER)).filter(name=OTHER),
            "rx_init_ago": None,
            "managed_by_other": None,
        }
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=htn_initial_review,
            model=HtnInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("managed_by_other", cm.exception.error_dict)

    def test_if_managed_by_drug_rx_init_after_dx(self):
        appointment = AppointmentMockModel()
        subject_visit = SubjectVisitMockModel(appointment)
        htn_initial_review = HtnInitialReviewMockModel()
        cleaned_data = {
            "subject_visit": subject_visit,
            "report_datetime": get_utcnow(),
            "dx_ago": "2y",
            "managed_by": MockSet(HtnTreatmentsMockModel(name=DRUGS)).filter(name=DRUGS),
            "rx_init_ago": "3y",
        }
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=htn_initial_review,
            model=HtnInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("rx_init_ago", cm.exception.error_dict)

        cleaned_data.update(rx_init_ago="2y")
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=htn_initial_review,
            model=HtnInitialReviewMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError:
            self.fail("ValidationError unexpectedly raised")

        cleaned_data.update(rx_init_ago="1y")
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=htn_initial_review,
            model=HtnInitialReviewMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError:
            self.fail("ValidationError unexpectedly raised")



from .consent_refusal_form_validator import ConsentRefusalFormValidator
from .health_facility_form_validator import HealthFacilityFormValidator
from .health_talk_log_form_validator import HealthTalkLogFormValidator
from .patient_call_form_validator import PatientCallFormValidator
from .patient_group_form_validator import PatientGroupFormValidator
from .patient_log_form_validator import (
    INVALID_APPOINTMENT_DATE,
    INVALID_CHANGE_ALREADY_SCREENED,
    INVALID_GROUP,
    PatientLogFormValidator,
)
from .subject_screening_form_validator import SubjectScreeningFormValidator



from django_mock_queries.query import MockModel


class SubjectScreeningMockModel(MockModel):
    def __init__(self, *args, **kwargs):
        kwargs["mock_name"] = "SubjectScreening"
        super().__init__(*args, **kwargs)
        self._meta.label_lower = "intecomm_screening.subjectscreening"

    def __str__(self):
        return str(self.screening_identifier)


class PatientGroupMockModel(MockModel):
    def __init__(self, *args, **kwargs):
        kwargs["mock_name"] = "PatientGroup"
        super().__init__(*args, **kwargs)
        self._meta.label_lower = "intecomm_screening.patientgroup"
        self.user_created = "frisco"
        self.user_modified = "frisco"

    def __str__(self):
        return str(self.name)

    def get_absolute_url(self) -> str:
        return "absolute_url"

    def get_changelist_url(self, search_term=None) -> str:
        return "changelist_url"

    def refresh_from_db(self):
        pass


class PatientLogMockModel(MockModel):
    def __init__(self, *args, **kwargs):
        kwargs["mock_name"] = "PatientLog"
        super().__init__(*args, **kwargs)
        self._meta.label_lower = "intecomm_screening.patientlog"

    def __str__(self):
        return str(self.name)

    def get_absolute_url(self) -> str:
        return "absolute_url"

    def get_changelist_url(self, search_term=None) -> str:
        return "changelist_url"


class ConditionsMockModel(MockModel):
    def __init__(self, *args, **kwargs):
        kwargs["mock_name"] = "Conditions"
        super().__init__(*args, **kwargs)
        self._meta.label_lower = "intecomm_list.condition"


class SocialHarmsMockModel(MockModel):
    def __init__(self, *args, **kwargs):
        kwargs["mock_name"] = "SocialHarms"
        super().__init__(*args, **kwargs)
        self._meta.label_lower = "intecomm_subject.socialharms"


class VitalsMockModel(MockModel):
    def __init__(self, *args, **kwargs):
        kwargs["mock_name"] = "Vitals"
        super().__init__(*args, **kwargs)
        self._meta.label_lower = "intecomm_subject.vitals"


class AppointmentMockModel(MockModel):
    def __init__(self, *args, **kwargs):
        kwargs["mock_name"] = "Appointment"
        super().__init__(*args, **kwargs)
        self.visit_schedule_name = "visit_schedule"
        self._meta.label_lower = "edc_appointment.appointment"


class SubjectVisitMockModel(MockModel):
    def __init__(self, appointment, *args, **kwargs):
        kwargs["mock_name"] = "SubjectVisit"
        super().__init__(*args, **kwargs)
        self.appointment = appointment
        self.visit_schedule_name = appointment.visit_schedule_name
        self.schedule_name = appointment.schedule_name
        self.report_datetime = appointment.appt_datetime
        self._meta.label_lower = "intecomm_subject.subjectvisit"


class DmInitialReviewMockModel(MockModel):
    def __init__(self, *args, **kwargs):
        kwargs["mock_name"] = "DmInitialReview"
        super().__init__(*args, **kwargs)
        self._meta.label_lower = "intecomm_subject.dminitialreview"


class HivInitialReviewMockModel(MockModel):
    def __init__(self, *args, **kwargs):
        kwargs["mock_name"] = "HivInitialReview"
        super().__init__(*args, **kwargs)
        self._meta.label_lower = "intecomm_subject.hivinitialreview"


class HtnInitialReviewMockModel(MockModel):
    def __init__(self, *args, **kwargs):
        kwargs["mock_name"] = "HtnInitialReview"
        super().__init__(*args, **kwargs)
        self._meta.label_lower = "intecomm_subject.htninitialreview"


class DmTreatmentsMockModel(MockModel):
    def __init__(self, *args, **kwargs):
        kwargs["mock_name"] = "DmTreatments"
        super().__init__(*args, **kwargs)
        self._meta.label_lower = "intecomm_lists.dmtreatments"


class HtnTreatmentsMockModel(MockModel):
    def __init__(self, *args, **kwargs):
        kwargs["mock_name"] = "HtnTreatments"
        super().__init__(*args, **kwargs)
        self._meta.label_lower = "intecomm_lists.htntreatments"



from edc_constants.constants import YES
from edc_crf.crf_form_validator_mixins import CrfFormValidatorMixin
from edc_dx_review.utils import raise_if_clinical_review_does_not_exist
from edc_form_validators import FormValidator


class SocialHarmsFormValidator(CrfFormValidatorMixin, FormValidator):
    def clean(self):
        raise_if_clinical_review_does_not_exist(self.cleaned_data.get("subject_visit"))
        for prefix in ["partner", "family", "friend", "coworker"]:
            self.applicable_if(YES, field=f"{prefix}", field_applicable=f"{prefix}_disclosure")

        for prefix in ["partner", "family", "friend", "coworker"]:
            self.applicable_if(
                YES, field=f"{prefix}_impact", field_applicable=f"{prefix}_impact_severity"
            )
            self.applicable_if(
                YES, field=f"{prefix}_impact", field_applicable=f"{prefix}_impact_status"
            )
            self.applicable_if(
                YES, field=f"{prefix}_impact", field_applicable=f"{prefix}_impact_help"
            )
            self.applicable_if(
                YES, field=f"{prefix}_impact", field_applicable=f"{prefix}_impact_referral"
            )

        for prefix in ["healthcare", "other_service", "employment", "insurance", "other"]:
            if prefix in ["other_service", "other"]:
                self.required_if(
                    YES,
                    field=f"{prefix}_impact",
                    field_required=f"{prefix}_impact_description",
                )
            self.applicable_if(
                YES, field=f"{prefix}_impact", field_applicable=f"{prefix}_impact_severity"
            )
            self.applicable_if(
                YES, field=f"{prefix}_impact", field_applicable=f"{prefix}_impact_status"
            )
            self.applicable_if(
                YES, field=f"{prefix}_impact", field_applicable=f"{prefix}_impact_help"
            )
            self.applicable_if(
                YES, field=f"{prefix}_impact", field_applicable=f"{prefix}_impact_referral"
            )



from __future__ import annotations

from datetime import datetime
from unittest.mock import patch

from dateutil.relativedelta import relativedelta
from dateutil.utils import today
from django import forms
from edc_constants.constants import COMPLETE, EQ, NO, NOT_APPLICABLE, YES
from edc_dx_review.constants import THIS_CLINIC
from edc_utils import get_utcnow

from intecomm_form_validators.subject import HivInitialReviewFormValidator as HivBase

from ..mock_models import HivInitialReviewMockModel
from ..test_case_mixin import TestCaseMixin


class HivInitialReviewTests(TestCaseMixin):
    def setUp(self) -> None:
        super().setUp()
        raise_missing_clinical_review_patcher = patch(
            "edc_dx_review.utils.raise_if_clinical_review_does_not_exist"
        )
        self.addCleanup(raise_missing_clinical_review_patcher.stop)
        self.raise_missing_clinical_review = raise_missing_clinical_review_patcher.start()

        self.dx_date = get_utcnow() - relativedelta(days=60)
        self.rx_date = self.dx_date + relativedelta(days=3)

    @staticmethod
    def get_form_validator_cls():
        class HivInitialReviewFormValidator(HivBase):
            pass

        return HivInitialReviewFormValidator

    def get_cleaned_data(self, report_datetime: datetime = None) -> dict:
        cleaned_data = dict(
            subject_visit=self.get_subject_visit(),
            report_datetime=report_datetime or get_utcnow(),
            dx_date=get_utcnow() - relativedelta(days=60),
            dx_ago="",
            receives_care=YES,
            clinic=THIS_CLINIC,
            clinic_other="",
            rx_init=YES,
            rx_init_date=self.rx_date,
            rx_init_ago="",
            has_vl=NO,
            drawn_date=None,
            vl=None,
            vl_quantifier=NOT_APPLICABLE,
            has_cd4=NO,
            cd4=None,
            cd4_date=None,
            crf_status=COMPLETE,
            crf_status_comments="",
        )
        return cleaned_data

    def test_cleaned_data_ok(self):
        hiv_initial_review = HivInitialReviewMockModel()
        cleaned_data = self.get_cleaned_data()
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_cannot_enter_dx_ago_and_exact_date(self):
        hiv_initial_review = HivInitialReviewMockModel()
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "dx_date": today(),
                "dx_ago": "2y",
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("dx_ago", cm.exception.error_dict)
        self.assertIn(
            "'Date conflict. Do not provide a response here if diagnosis date is available.'",
            str(cm.exception.error_dict.get("dx_ago")),
        )

    def test_dx_date_only_ok(self):
        hiv_initial_review = HivInitialReviewMockModel()
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "dx_date": self.dx_date,
                "dx_ago": "",
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_dx_ago_only_ok(self):
        hiv_initial_review = HivInitialReviewMockModel()
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "dx_date": None,
                "dx_ago": "1y",
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_dx_date_after_report_datetime_raises(self):
        hiv_initial_review = HivInitialReviewMockModel()
        report_datetime = get_utcnow() - relativedelta(days=30)
        cleaned_data = self.get_cleaned_data(report_datetime=report_datetime)
        cleaned_data.update(
            {
                "dx_date": report_datetime + relativedelta(days=1),
                "dx_ago": "",
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("dx_date", cm.exception.error_dict)
        self.assertIn(
            "Diagnosis date must be on or before ",
            str(cm.exception.error_dict.get("dx_date")),
        )

    def test_invalid_dx_ago_raises(self):
        hiv_initial_review = HivInitialReviewMockModel()
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "dx_date": None,
                "dx_ago": "1dx",
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("dx_ago", cm.exception.error_dict)
        self.assertIn(
            "Expected format ",
            str(cm.exception.error_dict.get("dx_ago")),
        )

    def test_rx_init_requires_rx_init_date(self):
        hiv_initial_review = HivInitialReviewMockModel()
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "rx_init": YES,
                "rx_init_date": None,
                "rx_init_ago": "",
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("Complete the treatment date", str(cm.exception))

        cleaned_data.update(
            {
                "rx_init": YES,
                "rx_init_date": None,
                "rx_init_ago": "14d",
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

        cleaned_data.update(
            {
                "rx_init": YES,
                "rx_init_date": get_utcnow() - relativedelta(days=7),
                "rx_init_ago": "",
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_rx_date_after_report_datetime_raises(self):
        hiv_initial_review = HivInitialReviewMockModel()
        report_datetime = get_utcnow() - relativedelta(days=30)
        cleaned_data = self.get_cleaned_data(report_datetime=report_datetime)
        cleaned_data.update(
            {
                "dx_date": report_datetime - relativedelta(days=7),
                "dx_ago": "",
                "rx_init_date": report_datetime + relativedelta(days=1),
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("rx_init_date", cm.exception.error_dict)
        self.assertIn(
            "Cannot be after report date",
            str(cm.exception.error_dict.get("rx_init_date")),
        )

    def test_invalid_rx_init_ago_raises(self):
        hiv_initial_review = HivInitialReviewMockModel()
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "rx_init": YES,
                "rx_init_date": None,
                "rx_init_ago": "1yxxx",
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("rx_init_ago", cm.exception.error_dict)
        self.assertIn(
            "Expected format ",
            str(cm.exception.error_dict.get("rx_init_ago")),
        )

    def test_rx_init_without_dx_date_raises(self):
        hiv_initial_review = HivInitialReviewMockModel()
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "dx_date": None,
                "dx_ago": "",
                "rx_init": YES,
                "rx_init_date": get_utcnow(),
                "rx_init_ago": "",
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("Complete the diagnosis date", str(cm.exception))

    def test_rx_init_and_rx_init_after_dx(self):
        hiv_initial_review = HivInitialReviewMockModel()
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "report_datetime": get_utcnow(),
                "dx_date": None,
                "dx_ago": "2y",
                "rx_init": YES,
                "rx_init_date": None,
                "rx_init_ago": "3y",
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("rx_init_ago", cm.exception.error_dict)

        cleaned_data.update(rx_init_ago="2y")
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

        cleaned_data.update(rx_init_ago="1y")
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError:
            self.fail("ValidationError unexpectedly raised")

    def test_rx_init_not_applicable_if_receives_care_no(self):
        hiv_initial_review = HivInitialReviewMockModel()
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "report_datetime": get_utcnow(),
                "dx_date": None,
                "dx_ago": "2y",
                "receives_care": NO,
                "clinic": NOT_APPLICABLE,
                "rx_init": YES,
                "rx_init_date": None,
                "rx_init_ago": "3y",
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("rx_init", cm.exception.error_dict)
        self.assertIn(
            "This field is not applicable.",
            str(cm.exception.error_dict.get("rx_init")),
        )

    def test_rx_init_applicable_if_receives_care_yes(self):
        hiv_initial_review = HivInitialReviewMockModel()
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "report_datetime": get_utcnow(),
                "dx_date": None,
                "dx_ago": "2y",
                "receives_care": YES,
                "clinic": THIS_CLINIC,
                "rx_init": NOT_APPLICABLE,
                "rx_init_date": None,
                "rx_init_ago": "",
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("rx_init", cm.exception.error_dict)
        self.assertIn(
            "This field is applicable.",
            str(cm.exception.error_dict.get("rx_init")),
        )

    def test_rx_init_no_without_date_ok(self):
        hiv_initial_review = HivInitialReviewMockModel()
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "report_datetime": get_utcnow(),
                "dx_date": None,
                "dx_ago": "2y",
                "receives_care": YES,
                "clinic": THIS_CLINIC,
                "rx_init": NO,
                "rx_init_date": None,
                "rx_init_ago": "",
                "has_vl": NOT_APPLICABLE,
                "has_cd4": NOT_APPLICABLE,
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_rx_init_yes_with_rx_init_date_ok(self):
        hiv_initial_review = HivInitialReviewMockModel()
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "report_datetime": get_utcnow(),
                "dx_date": None,
                "dx_ago": "2y",
                "receives_care": YES,
                "clinic": THIS_CLINIC,
                "rx_init": YES,
                "rx_init_date": get_utcnow() - relativedelta(years=1),
                "rx_init_ago": "",
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_rx_init_yes_with_rx_init_ago_ok(self):
        hiv_initial_review = HivInitialReviewMockModel()
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "report_datetime": get_utcnow(),
                "dx_date": None,
                "dx_ago": "2y",
                "receives_care": YES,
                "clinic": THIS_CLINIC,
                "rx_init": YES,
                "rx_init_date": None,
                "rx_init_ago": "1y6m",
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_rx_init_no_with_rx_init_date_raises(self):
        hiv_initial_review = HivInitialReviewMockModel()
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "report_datetime": get_utcnow(),
                "dx_date": None,
                "dx_ago": "2y",
                "rx_init": NO,
                "rx_init_date": get_utcnow() - relativedelta(years=1),
                "rx_init_ago": "",
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("rx_init_date", cm.exception.error_dict)
        self.assertIn(
            "This field is not required",
            str(cm.exception.error_dict.get("rx_init_date")),
        )

    def test_rx_init_no_with_rx_init_ago_raises(self):
        hiv_initial_review = HivInitialReviewMockModel()
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "report_datetime": get_utcnow(),
                "dx_date": None,
                "dx_ago": "2y",
                "rx_init": NO,
                "rx_init_date": None,
                "rx_init_ago": "1y",
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("rx_init_ago", cm.exception.error_dict)
        self.assertIn(
            "This field is not required",
            str(cm.exception.error_dict.get("rx_init_ago")),
        )

    def test_vl_drawn_date_before_hiv_dx_date_raises(self):
        hiv_initial_review = HivInitialReviewMockModel()
        cleaned_data = self.get_cleaned_data()
        report_datetime = get_utcnow()
        cleaned_data.update(
            {
                "report_datetime": get_utcnow(),
                "dx_date": None,
                "dx_ago": "2y",
                "has_vl": YES,
                "drawn_date": report_datetime.date() - relativedelta(years=2, days=1),
                "vl": 200,
                "vl_quantifier": EQ,
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("drawn_date", cm.exception.error_dict)
        self.assertIn(
            "Invalid. Cannot be before HIV diagnosis.",
            str(cm.exception.error_dict.get("drawn_date")),
        )

        cleaned_data.update({"drawn_date": report_datetime.date() - relativedelta(years=2)})
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

        cleaned_data.update(
            {
                "dx_date": report_datetime.date() - relativedelta(years=1),
                "dx_ago": "",
                "drawn_date": report_datetime.date() - relativedelta(years=1, days=1),
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("drawn_date", cm.exception.error_dict)
        self.assertIn(
            "Invalid. Cannot be before HIV diagnosis.",
            str(cm.exception.error_dict.get("drawn_date")),
        )

        cleaned_data.update({"drawn_date": report_datetime.date() - relativedelta(years=1)})
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_cd4_date_before_hiv_dx_date_raises(self):
        hiv_initial_review = HivInitialReviewMockModel()
        cleaned_data = self.get_cleaned_data()
        report_datetime = get_utcnow()
        cleaned_data.update(
            {
                "report_datetime": get_utcnow(),
                "dx_date": None,
                "dx_ago": "2y",
                "has_cd4": YES,
                "cd4": 100,
                "cd4_date": report_datetime.date() - relativedelta(years=2, days=1),
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("cd4_date", cm.exception.error_dict)
        self.assertIn(
            "Invalid. Cannot be before HIV diagnosis.",
            str(cm.exception.error_dict.get("cd4_date")),
        )

        cleaned_data.update({"cd4_date": report_datetime.date() - relativedelta(years=2)})
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

        cleaned_data.update(
            {
                "dx_date": report_datetime.date() - relativedelta(years=1),
                "dx_ago": "",
                "cd4_date": report_datetime.date() - relativedelta(years=1, days=1),
            }
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("cd4_date", cm.exception.error_dict)
        self.assertIn(
            "Invalid. Cannot be before HIV diagnosis.",
            str(cm.exception.error_dict.get("cd4_date")),
        )

        cleaned_data.update({"cd4_date": report_datetime.date() - relativedelta(years=1)})
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data,
            instance=hiv_initial_review,
            model=HivInitialReviewMockModel,
        )
        try:
            form_validator.validate()
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")
